"""Models and functions for CLI JSON schema dump."""
import argparse
import sys

from ywh2bt.cli.error import CliError
from ywh2bt.core.core import (
    AVAILABLE_SCHEMA_DUMP_FORMATS,
    write_message,
)
from ywh2bt.core.schema.error import SchemaError


def dump_schema(
    args: argparse.Namespace,
) -> None:
    """
    Dump a Schema of the structure of the configuration files.

    Args:
        args: command line arguments

    Raises:
        CliError: if an error occurred
    """
    dumper = AVAILABLE_SCHEMA_DUMP_FORMATS.get(args.output_format)
    if not dumper:
        raise CliError(f"Unsupported schema dump format {repr(args.output_format)}")
    try:
        schema = dumper()
    except SchemaError as schema_error:
        raise CliError("JSON Schema error") from schema_error
    write_message(
        message=schema,
        stream=sys.stdout,
    )



"""Models and functions used for data synchronisation between YesWeHack and trackers."""
from __future__ import annotations

from abc import (
    ABC,
    abstractmethod,
)
from dataclasses import dataclass
from string import Template
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Tuple,
    cast,
)

from ywh2bt.core.api.models.report import (
    REPORT_STATUS_TRANSLATIONS,
    AskForFixverificationStatusLog,
    Attachment,
    CloseLog,
    CommentLog,
    CvssUpdateLog,
    DetailsUpdateLog,
    FixVerifiedLog,
    Log,
    PriorityUpdateLog,
    Report,
    RewardLog,
    StatusUpdateLog,
    TrackerUpdateLog,
    TrackingStatusLog,
    TransferLog,
)
from ywh2bt.core.api.tracker import (
    SendLogsResult,
    TrackerAttachment,
    TrackerClient,
    TrackerClientError,
    TrackerIssue,
    TrackerIssueComment,
    TrackerIssueComments,
    TrackerIssueState,
)
from ywh2bt.core.api.yeswehack import (
    YesWeHackApiClient,
    YesWeHackApiClientError,
)
from ywh2bt.core.configuration.root import RootConfiguration
from ywh2bt.core.configuration.tracker import Trackers
from ywh2bt.core.configuration.yeswehack import (
    FeedbackOptions,
    Program,
    Programs,
    SynchronizeOptions,
    YesWeHackConfiguration,
    YesWeHackConfigurations,
)
from ywh2bt.core.factories.tracker_clients import TrackerClientsAbstractFactory
from ywh2bt.core.factories.yeswehack_api_clients import YesWeHackApiClientsAbstractFactory
from ywh2bt.core.markdown import markdown_to_ywh
from ywh2bt.core.state.decrypt import StateDecryptor
from ywh2bt.core.state.encrypt import StateEncryptor
from ywh2bt.core.synchronizer.error import SynchronizerError
from ywh2bt.core.synchronizer.listener import (
    NoOpSynchronizerListener,
    SynchronizerEndEvent,
    SynchronizerEndFetchReportsEvent,
    SynchronizerEndSendReportEvent,
    SynchronizerEvent,
    SynchronizerListener,
    SynchronizerStartEvent,
    SynchronizerStartFetchReportsEvent,
    SynchronizerStartSendReportEvent,
)
from ywh2bt.size import sizeof_fmt_si


class Synchronizer:
    """A class used for data synchronisation between YesWeHack and trackers."""

    _configuration: RootConfiguration
    _yes_we_hack_api_clients_factory: YesWeHackApiClientsAbstractFactory
    _tracker_clients_factory: TrackerClientsAbstractFactory
    _listener: SynchronizerListener
    _message_formatter: AbstractSynchronizerMessageFormatter

    def __init__(
        self,
        configuration: RootConfiguration,
        yes_we_hack_api_clients_factory: YesWeHackApiClientsAbstractFactory,
        tracker_clients_factory: TrackerClientsAbstractFactory,
        listener: Optional[SynchronizerListener] = None,
        message_formatter: Optional[AbstractSynchronizerMessageFormatter] = None,
    ):
        """
        Initialize self.

        Args:
            configuration: a configuration
            yes_we_hack_api_clients_factory: a YesWeHackApiClients factory
            tracker_clients_factory: a TrackerClients factory
            listener: an observer that will receive synchronization events
            message_formatter: a message formatter
        """
        self._configuration = configuration
        self._yes_we_hack_api_clients_factory = yes_we_hack_api_clients_factory
        self._tracker_clients_factory = tracker_clients_factory
        self._listener = listener or NoOpSynchronizerListener()
        self._message_formatter = message_formatter or SynchronizerMessageFormatter()

    def synchronize(
        self,
    ) -> None:
        """Synchronize YesWeHack and the trackers."""
        self._send_event(
            event=SynchronizerStartEvent(
                configuration=self._configuration,
            ),
        )
        yeswehack_configurations = cast(YesWeHackConfigurations, self._configuration.yeswehack)
        for yeswehack_name, yeswehack_configuration in yeswehack_configurations.items():
            self._send_programs_from_yeswehack_configuration(
                configuration=self._configuration,
                yeswehack_name=yeswehack_name,
                yeswehack_configuration=yeswehack_configuration,
            )
        self._send_event(
            event=SynchronizerEndEvent(
                configuration=self._configuration,
            ),
        )

    def _send_event(
        self,
        event: SynchronizerEvent,
    ) -> None:
        self._listener.on_event(
            event=event,
        )

    def _send_programs_from_yeswehack_configuration(
        self,
        configuration: RootConfiguration,
        yeswehack_name: str,
        yeswehack_configuration: YesWeHackConfiguration,
    ) -> None:
        yeswehack_client = self._yes_we_hack_api_clients_factory.get_yeswehack_api_client(
            configuration=yeswehack_configuration,
        )
        programs = cast(Programs, yeswehack_configuration.programs)
        for program in programs:
            if program.slug is None:
                continue
            self._send_event(
                event=SynchronizerStartFetchReportsEvent(
                    configuration=configuration,
                    yeswehack_name=yeswehack_name,
                    yeswehack_configuration=yeswehack_configuration,
                    program_slug=program.slug,
                ),
            )
            reports = self._get_afi_reports(
                yeswehack_client=yeswehack_client,
                program=program,
            )
            self._send_event(
                event=SynchronizerEndFetchReportsEvent(
                    configuration=configuration,
                    yeswehack_name=yeswehack_name,
                    yeswehack_configuration=yeswehack_configuration,
                    program_slug=program.slug,
                    reports=reports,
                ),
            )
            self._send_reports_to_trackers(
                configuration=configuration,
                yeswehack_name=yeswehack_name,
                yeswehack_configuration=yeswehack_configuration,
                yeswehack_client=yeswehack_client,
                program=program,
                reports=reports,
            )

    def _get_afi_reports(
        self,
        yeswehack_client: YesWeHackApiClient,
        program: Program,
    ) -> List[Report]:
        program_slug = cast(str, program.slug)
        filters = {
            "filter[trackingStatus][0]": "AFI",
        }
        synchronize_options = cast(SynchronizeOptions, program.synchronize_options)
        feedback_options = cast(FeedbackOptions, program.feedback_options)
        include_tracked = any(
            (
                synchronize_options.upload_private_comments,
                synchronize_options.upload_public_comments,
                synchronize_options.upload_cvss_updates,
                synchronize_options.upload_details_updates,
                synchronize_options.upload_priority_updates,
                synchronize_options.upload_rewards,
                synchronize_options.upload_status_updates,
                feedback_options.download_tracker_comments,
                feedback_options.issue_closed_to_report_afv,
            ),
        )
        if include_tracked:
            filters["filter[trackingStatus][1]"] = "T"
        try:
            return yeswehack_client.get_program_reports(
                slug=program_slug,
                filters=filters,
            )
        except YesWeHackApiClientError as e:
            raise SynchronizerError(
                f"Unable to get AFI/T reports for program {program_slug}",
            ) from e

    def _send_reports_to_trackers(
        self,
        configuration: RootConfiguration,
        yeswehack_name: str,
        yeswehack_configuration: YesWeHackConfiguration,
        yeswehack_client: YesWeHackApiClient,
        program: Program,
        reports: List[Report],
    ) -> None:
        for report in reports:
            self._send_report_to_trackers(
                configuration=configuration,
                yeswehack_name=yeswehack_name,
                yeswehack_configuration=yeswehack_configuration,
                yeswehack_client=yeswehack_client,
                program=program,
                report=report,
            )

    def _send_report_to_trackers(
        self,
        configuration: RootConfiguration,
        yeswehack_name: str,
        yeswehack_configuration: YesWeHackConfiguration,
        yeswehack_client: YesWeHackApiClient,
        program: Program,
        report: Report,
    ) -> None:
        if not program.bugtrackers_name:
            return
        for tracker_name in program.bugtrackers_name:
            self._send_report_to_tracker(
                configuration=configuration,
                yeswehack_name=yeswehack_name,
                yeswehack_configuration=yeswehack_configuration,
                yeswehack_client=yeswehack_client,
                tracker_name=tracker_name,
                synchronize_options=cast(SynchronizeOptions, program.synchronize_options),
                feedback_options=cast(FeedbackOptions, program.feedback_options),
                report=report,
            )

    def _send_report_to_tracker(
        self,
        configuration: RootConfiguration,
        yeswehack_name: str,
        yeswehack_configuration: YesWeHackConfiguration,
        yeswehack_client: YesWeHackApiClient,
        tracker_name: str,
        synchronize_options: SynchronizeOptions,
        feedback_options: FeedbackOptions,
        report: Report,
    ) -> None:
        tracker_configuration = cast(Trackers, configuration.bugtrackers)[tracker_name]
        tracker_client = self._tracker_clients_factory.get_tracker_client(
            configuration=tracker_configuration,
        )
        self._send_event(
            event=SynchronizerStartSendReportEvent(
                configuration=configuration,
                yeswehack_name=yeswehack_name,
                yeswehack_configuration=yeswehack_configuration,
                program_slug=report.program.slug,
                tracker_name=tracker_name,
                report=report,
            ),
        )
        report_synchronizer = ReportSynchronizer(
            report=report,
            yeswehack_client=yeswehack_client,
            tracker_name=tracker_name,
            tracker_client=tracker_client,
            synchronize_options=synchronize_options,
            feedback_options=feedback_options,
            message_formatter=self._message_formatter,
        )
        synchronize_report_result = report_synchronizer.synchronize_report()
        is_created_issue = synchronize_report_result.is_created_issue
        self._send_event(
            event=SynchronizerEndSendReportEvent(
                configuration=configuration,
                yeswehack_name=yeswehack_name,
                yeswehack_configuration=yeswehack_configuration,
                program_slug=report.program.slug,
                tracker_name=tracker_name,
                report=report,
                is_created_issue=is_created_issue,
                is_existing_issue=synchronize_report_result.is_existing_issue,
                new_report_status=synchronize_report_result.new_report_status,
                tracking_status_updated=report.tracking_status != "T" or not is_created_issue,
                tracker_issue=synchronize_report_result.send_logs_result.tracker_issue,
                issue_added_comments=[
                    comment.comment_id for comment in synchronize_report_result.send_logs_result.added_comments or []
                ],
                report_added_comments=synchronize_report_result.download_comments_result.downloaded_comments or [],
            ),
        )


class ReportSynchronizer:
    """A class used for data synchronisation between a YesWeHack report and a tracker issue."""

    _report: Report
    _yeswehack_client: YesWeHackApiClient
    _tracker_name: str
    _tracker_client: TrackerClient[Any]
    _synchronize_options: SynchronizeOptions
    _feedback_options: FeedbackOptions
    _message_formatter: AbstractSynchronizerMessageFormatter

    def __init__(
        self,
        report: Report,
        yeswehack_client: YesWeHackApiClient,
        tracker_name: str,
        tracker_client: TrackerClient[Any],
        synchronize_options: SynchronizeOptions,
        feedback_options: FeedbackOptions,
        message_formatter: Optional[AbstractSynchronizerMessageFormatter] = None,
    ) -> None:
        """
        Initialize self.

        Args:
            report: a report to be synchronized
            yeswehack_client: a YesWeHackApiClient
            tracker_name: a tracker name
            tracker_client: a TrackerClient
            synchronize_options: synchronization options
            feedback_options: feedback options
            message_formatter: an optional message formatter
        """
        self._report = report
        self._yeswehack_client = yeswehack_client
        self._tracker_name = tracker_name
        self._tracker_client = tracker_client
        self._synchronize_options = synchronize_options
        self._feedback_options = feedback_options
        self._message_formatter = message_formatter or SynchronizerMessageFormatter()

    def synchronize_report(
        self,
    ) -> SynchronizeReportResult:
        """
        Synchronize a YesWeHack report and a tracker issue.

        Raises:
            SynchronizerError: if a synchronization error occur

        Returns:
            the result of the synchronization
        """
        log = self._report.get_last_tracking_status_update_log(
            tracker_name=self._tracker_name,
        )
        is_created_issue = False
        tracker_issue = None
        if isinstance(log, TrackingStatusLog):
            is_created_issue = True
            tracker_issue = self._get_tracker_issue_from_logs(
                log=log,
            )
        if tracker_issue is None:
            if isinstance(log, TrackingStatusLog) and not self._synchronize_options.recreate_missing_issues:
                return SynchronizeReportResult(
                    is_created_issue=is_created_issue,
                    is_existing_issue=False,
                    new_report_status=None,
                    send_logs_result=SendLogsResult(
                        tracker_issue=None,
                        added_comments=[],
                    ),
                    download_comments_result=DownloadCommentsResult(
                        downloaded_comments=[],
                    ),
                )
            is_created_issue = False
            tracker_issue = self._create_tracker_issue()
        if not isinstance(tracker_issue, TrackerIssue):
            raise SynchronizerError(
                f"Unable to create new or get existing issue for #{self._report.report_id} in {self._tracker_name}",
            )
        if self._report.tracking_status != "T" or not is_created_issue:
            self._update_tracking_status(
                tracker_issue=tracker_issue,
            )
        tracker_issue_state = TrackerIssueState(
            closed=False,
            bugtracker_name=self._tracker_name,
        )
        logs = self._report.logs
        if is_created_issue:
            log_state = self._get_last_tracker_update_log()
            if log_state:
                last_tracker_update_log, tracker_issue_state = log_state
                if last_tracker_update_log:
                    logs = self._report.get_logs_after(
                        log=last_tracker_update_log,
                    )
        send_logs_result = self._send_synchronizable_logs(
            tracker_issue=tracker_issue,
            logs=logs,
        )
        for added_comment in send_logs_result.added_comments:
            tracker_issue_state.add_downloaded_comment(added_comment.comment_id)
        download_comments_result = self._download_comments(
            tracker_issue=tracker_issue,
            exclude_comments=tracker_issue_state.downloaded_comments,
        )
        for downloaded_comment in download_comments_result.downloaded_comments:
            tracker_issue_state.add_downloaded_comment(downloaded_comment)
        new_report_status = self._update_report_ask_for_fix_verification_status(
            tracker_issue=tracker_issue,
            tracker_issue_state=tracker_issue_state,
        )
        issue_status = self._get_issue_status_change(
            tracker_issue=tracker_issue,
            tracker_issue_state=tracker_issue_state,
        )
        tracker_issue_state.closed = tracker_issue.closed
        self._post_report_tracker_update(
            tracker_issue=tracker_issue,
            tracker_issue_state=tracker_issue_state,
            send_logs_result=send_logs_result,
            download_comments_result=download_comments_result,
            new_report_status=new_report_status,
            issue_status=issue_status,
        )
        return SynchronizeReportResult(
            is_created_issue=is_created_issue,
            is_existing_issue=True,
            new_report_status=new_report_status,
            send_logs_result=send_logs_result,
            download_comments_result=download_comments_result,
        )

    def _get_issue_status_change(
        self,
        tracker_issue: TrackerIssue,
        tracker_issue_state: TrackerIssueState,
    ) -> str:
        if tracker_issue.closed:
            return "Unchanged (closed)" if tracker_issue_state.closed else "Opened -> Closed"
        return "Closed -> Opened" if tracker_issue_state.closed else "Unchanged (opened)"

    def _get_tracker_issue_from_logs(
        self,
        log: TrackingStatusLog,
    ) -> Optional[TrackerIssue]:
        if all((log.tracker_id, log.tracker_url)):
            return self._tracker_client.get_tracker_issue(
                issue_id=cast(str, log.tracker_id),
            )
        return None

    def _create_tracker_issue(
        self,
    ) -> TrackerIssue:
        try:
            tracker_issue = self._tracker_client.send_report(
                report=self._report,
            )
        except TrackerClientError as send_error:
            raise SynchronizerError(
                f"Unable to send report #{self._report.report_id} to {self._tracker_name}",
            ) from send_error
        return tracker_issue

    def _update_tracking_status(
        self,
        tracker_issue: TrackerIssue,
    ) -> None:
        message = self._message_formatter.format_tracking_status_update_message(
            tracker_type=self._tracker_client.tracker_type,
            tracker_issue=tracker_issue,
        )
        try:
            self._yeswehack_client.put_report_tracking_status(
                report=self._report,
                status="T",
                tracker_name=self._tracker_name,
                issue_id=tracker_issue.issue_id,
                issue_url=tracker_issue.issue_url,
                comment=message,
            )
        except YesWeHackApiClientError as tracking_status_error:
            raise SynchronizerError(
                f"Unable to update tracking status for report #{self._report.report_id}",
            ) from tracking_status_error

    def _get_last_tracker_update_log(
        self,
    ) -> Optional[Tuple[Log, TrackerIssueState]]:
        for log in reversed(self._report.logs):
            if isinstance(log, TrackerUpdateLog):
                state = StateDecryptor.decrypt(
                    encrypted_state=log.tracker_token or "",
                    key=self._report.report_id,
                    state_type=TrackerIssueState,
                )
                if state and state.bugtracker_name == self._tracker_name:
                    return log, state
        return None

    def _send_synchronizable_logs(
        self,
        tracker_issue: TrackerIssue,
        logs: List[Log],
    ) -> SendLogsResult:
        synchronizable_logs = []
        for log in logs:
            synchronize = self._is_synchronizable_log(
                log=log,
            )
            if synchronize:
                synchronizable_logs.append(log)
        if not synchronizable_logs:
            return SendLogsResult(
                tracker_issue=tracker_issue,
                added_comments=[],
            )
        return self._send_logs(
            tracker_issue=tracker_issue,
            logs=synchronizable_logs,
        )

    def _is_synchronizable_log(
        self,
        log: Log,
    ) -> bool:
        synchronize_options = self._synchronize_options
        return any(
            (
                isinstance(log, CloseLog) and synchronize_options.upload_status_updates,
                isinstance(log, CommentLog) and synchronize_options.upload_public_comments and not log.private,
                (
                    isinstance(log, CommentLog)
                    and synchronize_options.upload_private_comments
                    and log.log_type == "comment"
                    and log.private
                ),
                (
                    isinstance(log, CommentLog)
                    and synchronize_options.upload_assign_comments
                    and log.log_type == "assign"
                    and log.private
                ),
                isinstance(log, CvssUpdateLog) and synchronize_options.upload_cvss_updates,
                isinstance(log, DetailsUpdateLog) and synchronize_options.upload_details_updates,
                isinstance(log, PriorityUpdateLog) and synchronize_options.upload_priority_updates,
                isinstance(log, RewardLog) and synchronize_options.upload_rewards,
                (
                    isinstance(log, (StatusUpdateLog, FixVerifiedLog, AskForFixverificationStatusLog))
                    and synchronize_options.upload_status_updates
                ),
                isinstance(log, TransferLog) and synchronize_options.upload_transfer_updates,
            )
        )

    def _send_logs(
        self,
        tracker_issue: TrackerIssue,
        logs: List[Log],
    ) -> SendLogsResult:
        try:
            return self._tracker_client.send_logs(
                tracker_issue=tracker_issue,
                logs=logs,
            )
        except TrackerClientError as send_error:
            raise SynchronizerError(
                f"Unable to send logs for #{self._report.report_id} to {self._tracker_name}",
            ) from send_error

    def _download_comments(
        self,
        tracker_issue: TrackerIssue,
        exclude_comments: Optional[List[str]] = None,
    ) -> DownloadCommentsResult:
        if not self._feedback_options.download_tracker_comments:
            return DownloadCommentsResult(
                downloaded_comments=[],
            )
        downloaded_comments = []
        comments = self._get_tracker_issue_comments(
            tracker_issue=tracker_issue,
            exclude_comments=exclude_comments,
        )
        for comment in comments:
            if comment.comment_id in (exclude_comments or []):
                continue
            self._download_comment(
                tracker_issue=tracker_issue,
                tracker_comment=comment,
            )
            downloaded_comments.append(comment.comment_id)
        return DownloadCommentsResult(
            downloaded_comments=downloaded_comments,
        )

    def _get_tracker_issue_comments(
        self,
        tracker_issue: TrackerIssue,
        exclude_comments: Optional[List[str]] = None,
    ) -> TrackerIssueComments:
        try:
            return self._tracker_client.get_tracker_issue_comments(
                issue_id=tracker_issue.issue_id,
                exclude_comments=exclude_comments,
            )
        except TrackerClientError as issue_comments_error:
            raise SynchronizerError(
                f"Unable to get issue comments for #{tracker_issue.issue_id}",
            ) from issue_comments_error

    def _download_comment(
        self,
        tracker_issue: TrackerIssue,
        tracker_comment: TrackerIssueComment,
    ) -> None:
        failed_attachments = {}
        attachments = {}
        for attachment_key, attachment in tracker_comment.attachments.items():
            try:
                uploaded_attachment = self._yeswehack_client.post_report_attachment(
                    report=self._report,
                    filename=attachment.filename,
                    file_content=attachment.content,
                    file_type=attachment.mime_type,
                )
            except YesWeHackApiClientError:
                failed_attachments[attachment_key] = attachment
                continue
            attachments[attachment_key] = uploaded_attachment
        try:
            self._yeswehack_client.post_report_tracker_message(
                report=self._report,
                tracker_name=self._tracker_name,
                issue_id=tracker_issue.issue_id,
                issue_url=tracker_issue.issue_url,
                comment=self._message_formatter.format_download_comment(
                    comment=tracker_comment,
                    attachments=attachments,
                    failed_attachments=failed_attachments,
                ),
                attachments=[uploaded_attachment.name for attachment_key, uploaded_attachment in attachments.items()],
            )
        except YesWeHackApiClientError as tracker_message_error:
            raise SynchronizerError(
                f"Unable to download comment {tracker_comment.comment_id} for report #{self._report.report_id}",
            ) from tracker_message_error

    def _post_report_tracker_update(
        self,
        tracker_issue: TrackerIssue,
        tracker_issue_state: TrackerIssueState,
        send_logs_result: SendLogsResult,
        download_comments_result: DownloadCommentsResult,
        new_report_status: Optional[Tuple[str, str]],
        issue_status: str,
    ) -> None:
        post_update = any(
            (
                send_logs_result.added_comments,
                download_comments_result.downloaded_comments,
                tracker_issue_state.changed,
            ),
        )
        if post_update:
            try:
                self._yeswehack_client.post_report_tracker_update(
                    report=self._report,
                    tracker_name=self._tracker_name,
                    issue_id=tracker_issue.issue_id,
                    issue_url=tracker_issue.issue_url,
                    token=StateEncryptor.encrypt(
                        key=self._report.report_id,
                        state=tracker_issue_state,
                    ),
                    comment=self._message_formatter.format_synchronization_done_message(
                        tracker_type=self._tracker_client.tracker_type,
                        report=self._report,
                        send_logs_result=send_logs_result,
                        download_comments_result=download_comments_result,
                        new_report_status=new_report_status,
                        issue_status=issue_status,
                    ),
                )
            except YesWeHackApiClientError as tracker_update_error:
                raise SynchronizerError(
                    f"Unable to post tracker update for report #{self._report.report_id}",
                ) from tracker_update_error

    def _update_report_status(
        self,
        tracker_issue: TrackerIssue,
        tracker_issue_state: TrackerIssueState,
    ) -> Optional[Tuple[str, str]]:
        if not self._feedback_options.issue_closed_to_report_afv:
            return None
        condition = tracker_issue.closed and not tracker_issue_state.closed
        ask_verif_status = "ask_verif"
        if condition and self._report.status != ask_verif_status:
            status_updated = self._put_report_status(
                status=ask_verif_status,
                comment="\n".join(
                    (
                        "Hello,",
                        "A fix has been deployed for this vulnerability.",
                        "Could you please verify that you cannot reproduce the bug nor bypass the fix?",
                        "Thanks for your help,",
                        "Regards,",
                    ),
                ),
            )
            return (self._report.status, ask_verif_status) if status_updated else None
        return None

    def _put_report_status(
        self,
        status: str,
        comment: str,
    ) -> bool:
        try:
            self._yeswehack_client.put_status(
                report=self._report,
                status=status,
                comment=self._message_formatter.format_status_update_comment(
                    comment=comment,
                ),
            )
        except YesWeHackApiClientError:
            return False
        return True

    def _update_report_ask_for_fix_verification_status(
        self,
        tracker_issue: TrackerIssue,
        tracker_issue_state: TrackerIssueState,
    ) -> Optional[Tuple[str, str]]:
        if not self._feedback_options.issue_closed_to_report_afv:
            return None
        condition = tracker_issue.closed and not tracker_issue_state.closed
        pending_status = "PENDING"
        if condition and self._report.ask_for_fix_verification_status != pending_status:
            status_updated = self._put_report_ask_for_fix_verification_status(
                status=pending_status,
                comment="\n".join(
                    (
                        "Hello,",
                        "A fix has been deployed for this vulnerability.",
                        "Could you please verify that you cannot reproduce the bug nor bypass the fix?",
                        "Thanks for your help,",
                        "Regards,",
                    ),
                ),
            )
            return (self._report.ask_for_fix_verification_status, pending_status) if status_updated else None
        return None

    def _put_report_ask_for_fix_verification_status(
        self,
        status: str,
        comment: str,
    ) -> bool:
        try:
            self._yeswehack_client.put_ask_for_fix_verification_status(
                report=self._report,
                status=status,
                comment=self._message_formatter.format_status_update_comment(
                    comment=comment,
                ),
            )
        except YesWeHackApiClientError:
            return False
        return True


class AbstractSynchronizerMessageFormatter(ABC):
    """Abstract message formatter for a synchronizer."""

    @abstractmethod
    def format_tracking_status_update_message(
        self,
        tracker_type: str,
        tracker_issue: TrackerIssue,
    ) -> str:
        """
        Format a tracking status update message.

        Args:
            tracker_type: a tracker type
            tracker_issue: a tracker issue
        """

    @abstractmethod
    def format_synchronization_done_message(
        self,
        report: Report,
        tracker_type: str,
        send_logs_result: SendLogsResult,
        download_comments_result: DownloadCommentsResult,
        new_report_status: Optional[Tuple[str, str]],
        issue_status: str,
    ) -> str:
        """
        Format a synchronization done message.

        Args:
            report: a report
            tracker_type: a tracker type
            send_logs_result: a result of the synchronization of the report logs with the issue
            download_comments_result: a result of the synchronization of the issue comments with the report
            new_report_status: new status of the report, if changed
            issue_status: status of the issue
        """

    @abstractmethod
    def format_download_comment(
        self,
        comment: TrackerIssueComment,
        attachments: Dict[str, Attachment],
        failed_attachments: Dict[str, TrackerAttachment],
    ) -> str:
        """
        Format a downloaded comment.

        Args:
            comment: a comment
            attachments: a dict of attachments
            failed_attachments: a dict of tracker attachments that failed to be uploaded
        """

    @abstractmethod
    def format_status_update_comment(
        self,
        comment: str,
    ) -> str:
        """
        Format a report status update comment.

        Args:
            comment: a comment
        """


class SynchronizerMessageFormatter(AbstractSynchronizerMessageFormatter):
    """Message formatter for a synchronizer."""

    _tracking_status_update_template: Template = Template(
        "Synchronized with bugtracker: ${tracker_url} on project : ${project}."
        + "\n"
        + "Tracked to [${tracker_type} #${issue_id}](${issue_url}).",
    )
    _synchronization_done_template: Template = Template(
        "Synchronized with bugtracker: ${tracker_url} on project : ${project}."
        + "\n"
        + "Tracked to [${tracker_type} #${issue_id}](${issue_url})."
        + "\n"
        + "Report comments added to issue: ${issue_added_comment_count}"
        + "\n"
        + "Issue comments added to report: ${report_added_comment_count}"
        + "\n"
        + "Report status: ${report_status}"
        + "\n"
        + "Issue status: ${issue_status}",
    )
    _download_comment_template: Template = Template(
        "Comment from bugtracker:" + "\n" + "Date: ${date}" + "\n" + "Author: ${author}" + "\n\n" + "${comment}",
    )
    _failed_attachments_template: Template = Template(
        "**YWH2BT note:**"
        + "\n*There was an issue while uploading the following attachments from the bugtracker*:"
        + "\n\n"
        + "${attachments}",
    )
    _failed_attachment_template: Template = Template(
        "- ${filename} (size=${size}, mime=${mimetype})",
    )
    _status_update_comment_template: Template = Template(
        "${comment}",
    )

    def format_tracking_status_update_message(
        self,
        tracker_type: str,
        tracker_issue: TrackerIssue,
    ) -> str:
        """
        Format a tracking status update message.

        Args:
            tracker_type: a tracker type
            tracker_issue: a tracker issue

        Returns:
            a formatted message
        """
        return self._tracking_status_update_template.substitute(
            tracker_type=tracker_type,
            tracker_url=tracker_issue.tracker_url,
            project=tracker_issue.project,
            issue_id=tracker_issue.issue_id,
            issue_url=tracker_issue.issue_url,
        )

    def format_synchronization_done_message(
        self,
        report: Report,
        tracker_type: str,
        send_logs_result: SendLogsResult,
        download_comments_result: DownloadCommentsResult,
        new_report_status: Optional[Tuple[str, str]],
        issue_status: str,
    ) -> str:
        """
        Format a synchronization done message.

        Args:
            report: a report
            tracker_type: a tracker type
            send_logs_result: a result of the synchronization of the report logs with the issue
            download_comments_result: a result of the synchronization of the issue comments with the report
            new_report_status: new status of the report, if changed
            issue_status: status of the issue

        Returns:
            a formatted message
        """
        tracker_issue = send_logs_result.tracker_issue
        report_status = "Unchanged"
        if new_report_status:
            unknown_status_translation = "Unknown"
            old_status, new_status = new_report_status
            old_status_translation = REPORT_STATUS_TRANSLATIONS.get(old_status, unknown_status_translation)
            new_status_translation = REPORT_STATUS_TRANSLATIONS.get(new_status, unknown_status_translation)
            report_status = f"{old_status_translation} -> {new_status_translation}"
        return self._synchronization_done_template.substitute(
            tracker_type=tracker_type,
            tracker_url=tracker_issue.tracker_url if tracker_issue else "",
            project=tracker_issue.project if tracker_issue else "",
            issue_id=tracker_issue.issue_id if tracker_issue else "",
            issue_url=tracker_issue.issue_url if tracker_issue else "",
            issue_added_comment_count=len(send_logs_result.added_comments),
            report_added_comment_count=len(download_comments_result.downloaded_comments),
            report_status=report_status,
            issue_status=issue_status,
        )

    def format_download_comment(
        self,
        comment: TrackerIssueComment,
        attachments: Dict[str, Attachment],
        failed_attachments: Dict[str, TrackerAttachment],
    ) -> str:
        """
        Format a downloaded comment.

        Args:
            comment: a comment
            attachments: a dict of attachments
            failed_attachments: a dict of tracker attachments that failed to be uploaded

        Returns:
            a formatted comment
        """
        formatted_comment = self._download_comment_template.substitute(
            date=comment.created_at,
            author=comment.author,
            comment=markdown_to_ywh(
                message=comment.body,
                attachments=attachments,
            ),
        )
        if failed_attachments:
            formatted_attachments = []
            for _attachment_key, failed_attachment in failed_attachments.items():
                formatted_attachments.append(
                    self._failed_attachment_template.substitute(
                        filename=failed_attachment.filename,
                        mimetype=failed_attachment.mime_type,
                        size=sizeof_fmt_si(len(failed_attachment.content), precision=2),
                    ),
                )
            formatted_failed_attachments = self._failed_attachments_template.substitute(
                attachments="\n".join(formatted_attachments),
            )
            formatted_comment = f"{formatted_comment}\n\n{formatted_failed_attachments}"
        return formatted_comment

    def format_status_update_comment(
        self,
        comment: str,
    ) -> str:
        """
        Format a report status update comment.

        Args:
            comment: a comment

        Returns:
            a formatted comment for a status update
        """
        return self._status_update_comment_template.substitute(
            comment=comment,
        )

    def _string_state(
        self,
        is_created_issue: bool,
        issue_state: TrackerIssueState,
        last_state: Optional[TrackerIssueState],
    ) -> str:
        if is_created_issue:
            return self._string_state_for_update(
                issue_state=issue_state,
                last_state=last_state,
            )
        return "created"

    def _string_state_for_update(
        self,
        issue_state: TrackerIssueState,
        last_state: Optional[TrackerIssueState],
    ) -> str:
        string_state = []
        if issue_state != last_state:
            if last_state:
                string_state.append("closed" if last_state.closed else "opened")
            else:
                string_state.append("???")
            string_state.append(" -> ")
            string_state.append("closed" if issue_state.closed else "opened")
        return "".join(string_state)


@dataclass
class DownloadCommentsResult:
    """A result of downloading comments from a tracker."""

    downloaded_comments: List[str]


@dataclass
class SynchronizeReportResult:
    """A result of synchronizing a report with a tracker."""

    is_created_issue: bool
    is_existing_issue: bool
    new_report_status: Optional[Tuple[str, str]]
    send_logs_result: SendLogsResult
    download_comments_result: DownloadCommentsResult



"""Models and functions used for root configuration GUI."""
from datetime import datetime
from io import StringIO
from typing import (
    Any,
    Optional,
    Tuple,
    cast,
)

from PySide6.QtCore import (
    QFileInfo,
    QSettings,
    Signal,
)
from PySide6.QtGui import (
    QFontDatabase,
    QFontMetrics,
    Qt,
)
from PySide6.QtWidgets import (
    QFileDialog,
    QMessageBox,
    QPlainTextEdit,
    QProgressBar,
    QScrollArea,
    QSplitter,
    QTabWidget,
    QVBoxLayout,
    QWidget,
)

from ywh2bt.core.configuration.root import RootConfiguration
from ywh2bt.core.core import (
    AVAILABLE_FORMATS,
    get_root_configuration_loader,
)
from ywh2bt.core.error import print_error
from ywh2bt.core.exceptions import CoreException
from ywh2bt.core.loader import RootConfigurationLoader
from ywh2bt.gui.dialog.error import ErrorDialogMixin
from ywh2bt.gui.dialog.file import FileFormatDialogFilters
from ywh2bt.gui.hashing import file_checksum
from ywh2bt.gui.widgets.attribute.attributes_container_widget import AttributesContainerWidget
from ywh2bt.gui.widgets.logs_widget import (
    LogEntry,
    LogsWidget,
    LogType,
)
from ywh2bt.gui.widgets.root_configuration_entry import (
    RootConfigurationEntry,
    RootConfigurationEntryFile,
)
from ywh2bt.gui.widgets.thread.synchronizer import SynchronizerThread
from ywh2bt.gui.widgets.thread.tester import TesterThread
from ywh2bt.gui.widgets.typing import as_signal_instance


class RootConfigurationWidget(QWidget, ErrorDialogMixin):
    """Root configuration GUI."""

    root_configuration_modified: Signal = Signal(RootConfigurationEntry)
    root_configuration_saved: Signal = Signal(RootConfigurationEntry)
    root_configuration_saved_as: Signal = Signal(str, str)

    _log_entry_available: Signal = Signal(LogEntry)

    _file_format_dialog_filters: FileFormatDialogFilters
    _settings: QSettings

    _entry: RootConfigurationEntry

    _tab_widget: QTabWidget
    _edit_tab_widget: AttributesContainerWidget
    _raw_widget: QPlainTextEdit

    _logs_widget: LogsWidget
    _progress_bar: QProgressBar
    _thread_running: bool

    def __init__(
        self,
        settings: QSettings,
        entry: RootConfigurationEntry,
        parent: Optional[QWidget] = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        Initialize self.

        Args:
            settings: application settings
            entry: a root configuration entry
            parent: a parent widget
            args: extra arguments
            kwargs: extra keyword arguments
        """
        super().__init__(
            parent,
            *args,
            **kwargs,
        )
        self._entry = entry
        self._settings = settings
        self._file_format_dialog_filters = FileFormatDialogFilters(
            formats=AVAILABLE_FORMATS,
        )
        self._init_ui()
        self._on_entry_changed()
        as_signal_instance(self._log_entry_available).connect(
            self._on_log_entry_available,
        )
        self._thread_running = False

    def set_entry(
        self,
        entry: RootConfigurationEntry,
    ) -> None:
        """
        Set the entry.

        Args:
            entry: an entry
        """
        self._entry = entry
        self._on_entry_changed()

    def _on_entry_changed(
        self,
        update_raw: bool = True,
    ) -> None:
        self._edit_tab_widget.set_container(
            container=self._entry.configuration,
        )
        self._edit_tab_widget.setEnabled(self._entry.configuration is not None)
        self._validate_entry()
        if update_raw:
            self._update_raw_widget()

    def _init_ui(
        self,
    ) -> None:
        layout = QVBoxLayout(self)

        self._tab_widget = QTabWidget(self)

        self._edit_tab_widget = self._create_edit_tab_widget()
        self._tab_widget.addTab(
            self._wrap_in_scroll_area(
                widget=self._edit_tab_widget,
            ),
            "Configuration",
        )
        raw_tab_widget = self._create_raw_tab_widget()
        self._tab_widget.addTab(
            raw_tab_widget,
            f"Raw ({self._entry.raw_format})",
        )

        self._logs_widget = LogsWidget(self)

        self._progress_bar = self._create_progress_bar()

        splitter = self._create_splitter()

        layout.addWidget(splitter)

    def _create_edit_tab_widget(
        self,
    ) -> AttributesContainerWidget:
        widget = AttributesContainerWidget(
            parent=self,
            container_class=RootConfiguration,
            container=self._entry.configuration,
        )
        as_signal_instance(widget.container_changed).connect(
            self._on_root_configuration_modified,
        )
        return widget

    def _create_progress_bar(
        self,
    ) -> QProgressBar:
        widget = QProgressBar(self)
        widget.setRange(0, 100)
        widget.setFixedHeight(15)
        return widget

    def _create_splitter(
        self,
    ) -> QSplitter:
        splitter = QSplitter(
            Qt.Orientation.Vertical,
            self,
        )
        splitter.addWidget(self._tab_widget)
        splitter.addWidget(self._logs_widget)
        splitter.addWidget(self._progress_bar)
        splitter.setCollapsible(0, False)
        splitter.setCollapsible(1, False)
        splitter.setCollapsible(2, False)
        splitter.setStretchFactor(0, 98)
        splitter.setStretchFactor(1, 1)
        splitter.setStretchFactor(2, 1)
        return splitter

    def _wrap_in_scroll_area(
        self,
        widget: QWidget,
    ) -> QScrollArea:
        scroll_area = QScrollArea(self)
        scroll_area.setWidget(widget)
        scroll_area.setWidgetResizable(True)
        return scroll_area

    def _on_root_configuration_modified(
        self,
        configuration: RootConfiguration,
    ) -> None:
        self._entry.configuration = configuration
        self._validate_entry()
        self._update_raw_widget()
        as_signal_instance(self.root_configuration_modified).emit(
            self._entry,
        )

    def _create_raw_tab_widget(
        self,
    ) -> QWidget:
        layout = QVBoxLayout()
        self._raw_widget = self._create_raw_widget()
        layout.addWidget(self._raw_widget, 1)

        widget = QWidget(self)
        widget.setLayout(layout)
        return widget

    def _create_raw_widget(
        self,
    ) -> QPlainTextEdit:
        widget = QPlainTextEdit(self)
        font = QFontDatabase.systemFont(QFontDatabase.SystemFont.FixedFont)
        font.setFixedPitch(True)
        metrics = QFontMetrics(font)
        widget.setFont(font)
        widget.setTabStopDistance(2 * metrics.horizontalAdvance(" "))
        as_signal_instance(widget.textChanged).connect(
            self._on_raw_changed,
        )
        return widget

    def _on_raw_changed(
        self,
    ) -> None:
        raw = self._raw_widget.toPlainText()
        try:
            self._entry.raw = raw
        except CoreException as e:
            self._log(
                log_type=LogType.error,
                message=_format_error(
                    error=e,
                ),
            )
        self._on_entry_changed(
            update_raw=False,
        )
        as_signal_instance(self.root_configuration_modified).emit(
            self._entry,
        )

    def _validate_entry(
        self,
    ) -> None:
        if not self._entry.configuration:
            return
        try:
            self._entry.configuration.validate()
        except CoreException as e:
            self._log(
                log_type=LogType.error,
                message=_format_error(
                    error=e,
                ),
            )
        else:
            self._log(
                log_type=LogType.success,
                message="Valid",
            )

    def _on_log_entry_available(
        self,
        log_entry: LogEntry,
    ) -> None:
        self._logs_widget.add_log_entry(
            entry=log_entry,
        )

    def _log(
        self,
        message: str,
        log_type: LogType = LogType.standard,
    ) -> None:
        self._logs_widget.add_log_entry(
            entry=LogEntry(
                date_time=datetime.now(),
                context=self._entry.name,
                message=message,
                log_type=log_type,
            ),
        )

    def _get_loader(
        self,
    ) -> RootConfigurationLoader:
        return get_root_configuration_loader(
            file_format=self._entry.raw_format,
        )

    def _update_raw_widget(
        self,
    ) -> None:
        plain_text = self._entry.raw
        self._raw_widget.blockSignals(True)
        cursor = self._raw_widget.textCursor()
        position = cursor.position()
        self._raw_widget.setPlainText(plain_text)
        cursor.setPosition(min(position, len(plain_text)))
        self._raw_widget.setTextCursor(cursor)
        self._raw_widget.blockSignals(False)

    def _get_last_opened_dir(
        self,
    ) -> str:
        last_opened_file = cast(str, self._settings.value("last_opened_file"))
        file_info = QFileInfo(last_opened_file) if last_opened_file else QFileInfo()
        return file_info.dir().path()

    def save(
        self,
    ) -> None:
        """Engage the saving process."""
        entry = self._entry
        save_invalid = self._prompt_save_invalid_entry(
            entry=entry,
        )
        if not save_invalid:
            return
        path_format = self._prompt_save_path_format(
            entry=entry,
        )
        if not path_format:
            return
        file_path, file_format = path_format
        if entry.configuration is None:
            entry.configuration = RootConfiguration()
        saved = self._save_to_file(
            entry=entry,
            file_path=file_path,
            file_format=file_format,
        )
        if not saved:
            return
        self._setup_entry_from_saved_file(
            file_info=QFileInfo(file_path),
            file_format=file_format,
        )
        self._log(
            log_type=LogType.success,
            message=f"{self._entry.name} saved",
        )

    def save_as(
        self,
    ) -> None:
        """Engage the saving process into a new file."""
        entry = self._entry
        save_invalid = self._prompt_save_invalid_entry(
            entry=entry,
        )
        if not save_invalid:
            return
        path_format = self._prompt_new_file(
            format_filters=self._file_format_dialog_filters.get_filters_string(),
        )
        if not path_format:
            return
        file_path, file_format = path_format
        if entry.configuration is None:
            entry.configuration = RootConfiguration()
        saved = self._save_to_file(
            entry=entry,
            file_path=file_path,
            file_format=file_format,
        )
        if not saved:
            return
        as_signal_instance(self.root_configuration_saved_as).emit(
            file_path,
            file_format,
        )

    def _setup_entry_from_saved_file(
        self,
        file_info: QFileInfo,
        file_format: str,
    ) -> None:
        self._entry.file = RootConfigurationEntryFile(
            info=file_info,
            file_format=file_format,
            checksum=file_checksum(file_info.filePath()),
        )
        self._entry.name = file_info.fileName()
        self._entry.original_raw = self._entry.raw
        as_signal_instance(self.root_configuration_modified).emit(
            self._entry,
        )
        as_signal_instance(self.root_configuration_saved).emit(
            self._entry,
        )

    def test(
        self,
    ) -> None:
        """Test the configuration."""
        entry = self._entry
        if self._thread_running or not entry.configuration or not entry.configuration.is_valid():
            return
        thread = TesterThread(
            parent=self,
            entry=entry,
            log_entry_available_signal=as_signal_instance(self._log_entry_available),
        )
        as_signal_instance(thread.test_started).connect(
            self._on_thread_started,
        )
        as_signal_instance(thread.test_ended).connect(
            self._on_thread_ended,
        )
        thread.start()

    def synchronize(
        self,
    ) -> None:
        """Synchronize the configuration."""
        entry = self._entry
        if self._thread_running or not entry.configuration or not entry.configuration.is_valid():
            return
        thread = SynchronizerThread(
            parent=self,
            entry=entry,
            log_entry_available_signal=as_signal_instance(self._log_entry_available),
        )
        as_signal_instance(thread.synchronization_started).connect(
            self._on_thread_started,
        )
        as_signal_instance(thread.synchronization_ended).connect(
            self._on_thread_ended,
        )
        thread.start()

    def _on_thread_started(
        self,
    ) -> None:
        self._thread_running = True
        self._progress_bar.setRange(0, 0)

    def _on_thread_ended(
        self,
    ) -> None:
        self._progress_bar.setRange(0, 100)
        self._progress_bar.reset()
        self._thread_running = False

    def _prompt_save_invalid_entry(
        self,
        entry: RootConfigurationEntry,
    ) -> bool:
        error_message = None
        if entry.configuration is None:
            error_message = "Are you sure you want to save an empty configuration?"
        elif not entry.configuration.is_valid():
            error_message = "Are you sure you want to save an invalid configuration?"
        if error_message:
            reply = QMessageBox.question(
                self,
                f"Save {entry.name}",
                error_message,
                QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
                QMessageBox.StandardButton.No,
            )
            if reply == QMessageBox.StandardButton.No:
                return False
        return True

    def _prompt_save_path_format(
        self,
        entry: RootConfigurationEntry,
    ) -> Optional[Tuple[str, str]]:
        if entry.file:
            exists = entry.file.info.exists()
            changed = entry.file.checksum != file_checksum(entry.file.info.filePath())
            if exists and changed:
                overwrite = self._prompt_overwrite_changed_file(
                    entry=entry,
                )
                if not overwrite:
                    return None
            file_path = entry.file.info.filePath()
            file_format = entry.file.file_format
        else:
            new_file = self._prompt_new_file(
                format_filters=self._file_format_dialog_filters.get_filters_string_for(
                    format_name=self._entry.raw_format,
                )
                or "",
            )
            if not new_file:
                return None
            file_path, file_format = new_file
        return file_path, file_format

    def _prompt_overwrite_changed_file(
        self,
        entry: RootConfigurationEntry,
    ) -> bool:
        if not entry.file:
            return False
        reply = QMessageBox.question(
            self,
            f"Overwrite {entry.name}",
            f"{entry.file.info.filePath()} has changed on the disk.\nOverwrite?",
            QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
            QMessageBox.StandardButton.No,
        )
        return reply == QMessageBox.StandardButton.Yes

    def _prompt_new_file(
        self,
        format_filters: str,
    ) -> Optional[Tuple[str, str]]:
        last_opened_dir = self._get_last_opened_dir()
        file_path, chosen_filter = QFileDialog.getSaveFileName(
            self,
            "New file",
            last_opened_dir,
            format_filters or "",
        )
        if not file_path or not chosen_filter:
            return None
        file_format = self._file_format_dialog_filters.get_format_name(
            filter_string=chosen_filter,
        )
        if not file_format:
            return None
        return file_path, file_format

    def _save_to_file(
        self,
        entry: RootConfigurationEntry,
        file_path: str,
        file_format: str,
    ) -> bool:
        if not entry.configuration:
            return False
        loader = self._get_root_configuration_loader(
            file_format=file_format,
        )
        if not loader:
            return False
        try:
            with open(file_path, "w") as destination_file:
                loader.save(
                    data=entry.configuration,
                    stream=destination_file,
                )
        except (IOError, CoreException) as save_error:
            self.show_exception_dialog(
                parent=self,
                text="Save error",
                informative_text=f"Unable to save file:\n{file_path}",
                exception=save_error,
            )
            return False
        return True

    def _get_root_configuration_loader(
        self,
        file_format: str,
    ) -> Optional[RootConfigurationLoader]:
        try:
            return get_root_configuration_loader(
                file_format=file_format,
            )
        except CoreException as loader_error:
            self.show_exception_dialog(
                parent=self,
                text="Loader error",
                informative_text=f"Unable to get loader for {file_format}",
                exception=loader_error,
            )
        return None


def _format_error(
    error: Exception,
) -> str:
    error_stream = StringIO()
    print_error(
        stream=error_stream,
        error=error,
    )
    return error_stream.getvalue()



"""Models and functions used for formatting reports data."""
from string import Template
from typing import (
    Dict,
    Tuple,
    cast,
)

from bs4 import (  # type: ignore
    BeautifulSoup,
    Tag,
)

from ywh2bt.core.api.formatter.formatter import ReportMessageFormatter
from ywh2bt.core.converter.html2jira import html2jira


REPORT_TITLE_TEMPLATE = "${local_id} : ${title}"
REPORT_DESCRIPTION_TEMPLATE = """
|| Title || ${local_id} : ${title} ||
| Priority | ${priority_name} |
| ${bug_type_label} | [${bug_type_name}|${bug_type_link}] => [Remediation|${bug_type_remediation_link}] |
| ${scope_label} | {noformat}${scope}{noformat} |
| Severity | ${cvss_criticity}, score: ${cvss_score}, vector: {noformat}${cvss_vector}{noformat} |
| ${end_point_label} | ${end_point} |
| ${vulnerable_part_label} | ${vulnerable_part} |
| ${part_name_label} | ${part_name} |
| ${payload_sample_label} | ${payload_sample} |
| Technical Environment | ${technical_environment} |

${description}
"""
CLOSE_LOG_TEMPLATE = """
*Report closed*:

${old_status} -> ${new_status}

*Comment*:

${comment}
"""
COMMENT_BODY_TEMPLATE = """
*Date*: ${date}

${body}
"""
COMMENT_LOG_TEMPLATE = """
*Comment*:

${comment}
"""
CVSS_UPDATE_LOG_TEMPLATE = """
*CVSS updated*:

|| Detail || Old value || New value ||
| *Severity* | ${old_cvss_criticity} | ${new_cvss_criticity} |
| *Score* | ${old_cvss_score} | ${new_cvss_score} |
"""
STATUS_UPDATE_LOG_TEMPLATE = """
*Status updated*:

${old_status} -> ${new_status}

*Comment*:

${comment}
"""
FIX_VERIFIED_LOG_TEMPLATE = """
*Fix verification*:

${verified}

*Comment*:

${comment}
"""
DETAILS_UPDATE_LOG_TEMPLATE = """
*Details updated*:

|| Detail || Old value || New value ||
${details_lines}
"""
DETAILS_UPDATE_LOG_LINE_TEMPLATE = """| *${updated_property}* | {noformat}${old_value}{noformat} | {noformat}${new_value}{noformat} |
"""  # noqa: E501
PRIORITY_UPDATE_LOG_TEMPLATE = """
*Priority updated*:

${new_priority}
"""
REWARD_LOG_TEMPLATE = """
*Comment from reward*:

${comment}
"""
TRANSFER_LOG_TEMPLATE = """
*Report transferred*:

${old_program} -> ${program}
"""
ASK_FOR_FIX_VERIFICATION_STATUS_LOG_TEMPLATE = """
*Fix verification*:

${new_ask_for_fix_verification_status}

*Comment*:

${comment}
"""


class JiraReportMessageFormatter(ReportMessageFormatter):
    """A report formatter to JIRA."""

    _code_block_template = Template("{code${language}}\n${content}\n{code}")

    _code_block_languages = [
        "actionscript",
        "ada",
        "applescript",
        "bash",
        "c",
        "c#",
        "c++",
        "cpp",
        "css",
        "erlang",
        "go",
        "groovy",
        "haskell",
        "html",
        "java",
        "javascript",
        "js",
        "json",
        "lua",
        "none",
        "nyan",
        "objc",
        "perl",
        "php",
        "python",
        "r",
        "rainbow",
        "ruby",
        "scala",
        "sh",
        "sql",
        "swift",
        "visualbasic",
        "xml",
        "yaml",
    ]

    def __init__(self) -> None:
        """Initialize self."""
        super().__init__(
            report_title_template=Template(REPORT_TITLE_TEMPLATE),
            report_description_template=Template(REPORT_DESCRIPTION_TEMPLATE),
            close_log_template=Template(CLOSE_LOG_TEMPLATE),
            comment_body_template=Template(COMMENT_BODY_TEMPLATE),
            comment_log_template=Template(COMMENT_LOG_TEMPLATE),
            cvss_update_log_template=Template(CVSS_UPDATE_LOG_TEMPLATE),
            status_update_log_template=Template(STATUS_UPDATE_LOG_TEMPLATE),
            fix_verified_log_template=Template(FIX_VERIFIED_LOG_TEMPLATE),
            details_update_log_template=Template(DETAILS_UPDATE_LOG_TEMPLATE),
            details_update_log_line_template=Template(DETAILS_UPDATE_LOG_LINE_TEMPLATE),
            priority_update_log_template=Template(PRIORITY_UPDATE_LOG_TEMPLATE),
            reward_log_template=Template(REWARD_LOG_TEMPLATE),
            transfer_log_template=Template(TRANSFER_LOG_TEMPLATE),
            ask_for_fix_verification_status_log_template=Template(ASK_FOR_FIX_VERIFICATION_STATUS_LOG_TEMPLATE),
        )

    def transform_report_description_html(
        self,
        description_html: str,
    ) -> str:
        """
        Transform the report description.

        Args:
            description_html: a report description in HTML

        Returns:
            a transformed description
        """
        return self._html2jira(
            html=description_html,
        )

    def transform_html(
        self,
        html: str,
    ) -> str:
        """
        Transform from HTML.

        Args:
            html: HTML

        Returns:
            HTML transformed to JIRA
        """
        return self._html2jira(
            html=html,
        )

    def _html2jira(
        self,
        html: str,
    ) -> str:
        body_html = self._img_tags_to_jira_tags(
            html=html,
        )
        body_html, code_blocks = self._extract_code_blocks(
            html=body_html,
        )
        body_jira = html2jira(
            html=body_html,
        )
        for block_id, code_block in code_blocks.items():
            body_jira = body_jira.replace(
                block_id,
                code_block,
            )
        return body_jira

    def _img_tags_to_jira_tags(
        self,
        html: str,
    ) -> str:
        soup = BeautifulSoup(
            html,
            features="lxml",
        )
        clean_html = str(soup)
        for img in soup.findAll("img"):
            alt = img.attrs.get("alt", "")
            src = img.attrs.get("src", "")
            a = f"!{alt}|{src}!"
            clean_html = clean_html.replace(str(img), a)
        return clean_html

    def _extract_code_blocks(
        self,
        html: str,
    ) -> Tuple[str, Dict[str, str]]:
        return self._extract_code_blocks_from_soup(
            soup=BeautifulSoup(
                html,
                features="lxml",
            ),
        )

    def _extract_code_blocks_from_soup(
        self,
        soup: BeautifulSoup,
    ) -> Tuple[str, Dict[str, str]]:
        clean_html = str(soup)
        code_blocks = {}
        for idx, code_tag in enumerate(soup.findAll("code")):
            block_id = f"{{yeswehack_code_section_{idx}}}"
            language = self._extract_code_block_language(
                code_tag=code_tag,
            )
            code_blocks[block_id] = self._code_block_template.substitute(
                language=f":{language}" if language else "",
                content="".join(map(str, code_tag.contents)),
            )
            clean_html = clean_html.replace(str(code_tag), block_id)
        return clean_html, code_blocks

    def _extract_code_block_language(
        self,
        code_tag: Tag,
    ) -> str:
        title = code_tag.attrs.get("class", "")
        if isinstance(title, list):
            title = title[0] if title else ""
        language = cast(str, title.lower().replace("language-", ""))
        return language if language in self._code_block_languages else ""



"""Models and functions used for tracker clients factories."""
from abc import (
    ABC,
    abstractmethod,
)
from typing import (
    Any,
    Dict,
    Optional,
    Type,
)

from ywh2bt.core.api.tracker import (
    TrackerClient,
    TrackerClientError,
)
from ywh2bt.core.api.trackers.github.tracker import GitHubTrackerClient
from ywh2bt.core.api.trackers.gitlab import GitLabTrackerClient
from ywh2bt.core.api.trackers.jira.tracker import JiraTrackerClient
from ywh2bt.core.api.trackers.servicenow.tracker import ServiceNowTrackerClient
from ywh2bt.core.configuration.tracker import TrackerConfiguration
from ywh2bt.core.configuration.trackers.github import GitHubConfiguration
from ywh2bt.core.configuration.trackers.gitlab import GitLabConfiguration
from ywh2bt.core.configuration.trackers.jira import JiraConfiguration
from ywh2bt.core.configuration.trackers.servicenow import ServiceNowConfiguration
from ywh2bt.core.exceptions import CoreException


class TrackerClientsAbstractFactory(ABC):
    """Abstract factory for TrackerClients."""

    @abstractmethod
    def get_tracker_client(
        self,
        configuration: TrackerConfiguration,
    ) -> TrackerClient[Any]:
        """
        Get the api client for the given configuration.

        Args:
            configuration: a configuration
        """


class TrackerClientsFactory(TrackerClientsAbstractFactory):
    """Concrete factory for TrackerClients."""

    _tracker_clients: Dict[TrackerConfiguration, TrackerClient[Any]]

    def __init__(
        self,
    ) -> None:
        """Initialize self."""
        self._tracker_clients = {}

    def get_tracker_client(
        self,
        configuration: TrackerConfiguration,
    ) -> TrackerClient[Any]:
        """
        Get the api client for the given configuration.

        Args:
            configuration: a configuration

        Returns:
            The client
        """
        if configuration not in self._tracker_clients:
            self._tracker_clients[configuration] = self._build_tracker_client(
                configuration=configuration,
            )
        return self._tracker_clients[configuration]

    def _build_tracker_client(
        self,
        configuration: TrackerConfiguration,
    ) -> TrackerClient[Any]:
        configuration_class = configuration.__class__
        tracker_client_class = TrackerClientClassesRegistry.get_tracker_client_class(
            configuration_class=configuration_class,
        )
        if not tracker_client_class:
            message = (
                f"{configuration_class} does not have an associated tracker client ; "
                + f"use {TrackerClientClassesRegistry.__name__}.register_tracker_client_class() to register."
            )
            raise CoreException(message)
        try:
            client = tracker_client_class(
                configuration=configuration,
            )
        except TrackerClientError as e:
            raise CoreException(f"Unable to create {configuration_class} API client") from e
        return client


TrackerConfigurationType = Type[TrackerConfiguration]
TrackerClientType = Type[TrackerClient[Any]]
TrackerClientClassesType = Dict[
    TrackerConfigurationType,
    TrackerClientType,
]


class TrackerClientClassesRegistry:
    """Registry for associating TrackerConfiguration classes with TrackerClient classes."""

    _classes: TrackerClientClassesType = {}

    @classmethod
    def register_tracker_client_class(
        cls,
        configuration_class: TrackerConfigurationType,
        client_class: TrackerClientType,
    ) -> None:
        """
        Associate a TrackerConfiguration class with a TrackerClient client class.

        Args:
            configuration_class: a TrackerConfiguration class
            client_class: a TrackerClient client class
        """
        cls._classes[configuration_class] = client_class

    @classmethod
    def unregister_tracker_client_class(
        cls,
        configuration_class: TrackerConfigurationType,
    ) -> None:
        """
        Dissociate a TrackerConfiguration class from a TrackerClient client class.

        Args:
            configuration_class: a TrackerConfiguration class
        """
        if configuration_class in cls._classes:
            cls._classes.pop(configuration_class)

    @classmethod
    def get_tracker_client_class(
        cls,
        configuration_class: TrackerConfigurationType,
    ) -> Optional[TrackerClientType]:
        """
        Get the TrackerConfiguration class associated with a TrackerClient client class.

        Args:
            configuration_class: a TrackerConfiguration class

        Returns:
            the TrackerClient client class if one is associated with the TrackerConfiguration class ; else None
        """
        return cls._classes.get(configuration_class)


TrackerClientClassesRegistry.register_tracker_client_class(
    configuration_class=GitHubConfiguration,
    client_class=GitHubTrackerClient,
)
TrackerClientClassesRegistry.register_tracker_client_class(
    configuration_class=GitLabConfiguration,
    client_class=GitLabTrackerClient,
)
TrackerClientClassesRegistry.register_tracker_client_class(
    configuration_class=JiraConfiguration,
    client_class=JiraTrackerClient,
)
TrackerClientClassesRegistry.register_tracker_client_class(
    configuration_class=ServiceNowConfiguration,
    client_class=ServiceNowTrackerClient,
)



"""Models and functions used for data synchronisation between YesWeHack and trackers."""
from __future__ import annotations

from abc import (
    ABC,
    abstractmethod,
)
from dataclasses import dataclass
from datetime import datetime
from typing import (
    Any,
    Dict,
    Generic,
    List,
    Optional,
    TypeVar,
)

from ywh2bt.core.api.client import TestableApiClient
from ywh2bt.core.api.models.report import (
    Log,
    Report,
)
from ywh2bt.core.configuration.tracker import TrackerConfiguration
from ywh2bt.core.exceptions import CoreException
from ywh2bt.core.state.state import State


class TrackerClientError(CoreException):
    """A base tracker client error."""


class TrackerIssueState(State):
    """A state of an issue."""

    _changed: bool = False

    _closed: bool = False
    _bugtracker_name: Optional[str] = None
    _downloaded_comments: Optional[List[str]] = None

    def __init__(
        self,
        closed: bool = False,
        bugtracker_name: Optional[str] = None,
        downloaded_comments: Optional[List[str]] = None,
    ):
        """
        Initialize self.

        Args:
            closed: a flag indicating the closed status of the tracker issue to which this state belongs
            bugtracker_name: a bug tracker name of the tracker issue to which this state belongs
            downloaded_comments:
            a list of comment ids that have been downloaded (from the bug tracker to yeswehack platform)
        """
        self._closed = closed
        self._bugtracker_name = bugtracker_name
        self._downloaded_comments = downloaded_comments

    def as_dict(self) -> Dict[str, Any]:
        """
        Get the fields as a new dict.

        Returns:
            the fields as a new dict
        """
        return {
            "closed": self._closed,
            "bugtracker_name": self._bugtracker_name,
            "downloaded_comments": list(set(self._downloaded_comments)) if self._downloaded_comments else None,
        }

    @property
    def changed(self) -> bool:
        """
        Indicate if the underlying values of the state have changed.

        Returns:
            True if the underlying values of the state have changed
        """
        return self._changed

    @property
    def closed(self) -> bool:
        """
        Indicate the closed status of the tracker issue to which this state belongs.

        Returns:
            True if the tracker issue is closed
        """
        return self._closed

    @closed.setter
    def closed(
        self,
        closed: bool,
    ) -> None:
        """
        Set the closed status of the tracker issue to which this state belongs.

        Args:
            closed: True if the tracker issue is closed
        """
        if self._closed == closed:
            return
        self._changed = True
        self._closed = closed

    @property
    def bugtracker_name(self) -> Optional[str]:
        """
        Get the bug tracker name of the tracker issue to which this state belongs.

        Returns:
            the bug tracker name
        """
        return self._bugtracker_name

    @bugtracker_name.setter
    def bugtracker_name(
        self,
        bugtracker_name: str,
    ) -> None:
        """
        Set the bug tracker name of the tracker issue to which this state belongs.

        Args:
            bugtracker_name: the bug tracker name
        """
        if self._bugtracker_name == bugtracker_name:
            return
        self._changed = True
        self._bugtracker_name = bugtracker_name

    @property
    def downloaded_comments(self) -> Optional[List[str]]:
        """
        Get the list of comment ids that have been downloaded (from the bug tracker to yeswehack platform).

        Returns:
            the list of comment ids
        """
        return self._downloaded_comments

    @downloaded_comments.setter
    def downloaded_comments(
        self,
        downloaded_comments: Optional[List[str]],
    ) -> None:
        """
        Set the list of comment ids that have been downloaded (from the bug tracker to yeswehack platform).

        Args:
            downloaded_comments: the list of comment ids
        """
        if self._downloaded_comments == downloaded_comments:
            return
        self._changed = True
        self._downloaded_comments = downloaded_comments

    def add_downloaded_comment(
        self,
        comment_id: str,
    ) -> None:
        """
        Add a downloaded comment to the state.

        Args:
            comment_id: a comment id
        """
        if not self.downloaded_comments:
            self.downloaded_comments = []
        self.downloaded_comments.append(comment_id)

    def __eq__(
        self,
        other: Optional[object],
    ) -> bool:
        """
        Indicate if 2 states are the same.

        Args:
            other: a state

        Returns:
            True if the states are the same
        """
        return isinstance(other, TrackerIssueState) and all(
            (
                self._closed == other.closed,
                self._bugtracker_name == other.bugtracker_name,
                self._downloaded_comments == other.downloaded_comments,
            ),
        )


@dataclass
class TrackerIssue:
    """Information about a tracker issue."""

    tracker_url: str
    project: str
    issue_id: str
    issue_url: str
    closed: bool


@dataclass
class TrackerIssueComment:
    """Information about a tracker comment on an issue."""

    attachments: Dict[str, TrackerAttachment]
    author: str
    comment_id: str
    created_at: datetime
    body: str


TrackerIssueComments = List[TrackerIssueComment]


@dataclass
class TrackerAttachment:
    """Attachment on a tracker issue or comment."""

    filename: str
    mime_type: str
    content: bytes


@dataclass
class SendLogsResult:
    """A result of sending logs to a tracker."""

    tracker_issue: Optional[TrackerIssue]
    added_comments: TrackerIssueComments


T = TypeVar("T", bound=TrackerConfiguration)


class TrackerClient(TestableApiClient, ABC, Generic[T]):
    """An abstract base tracker client."""

    configuration: T

    def __init__(
        self,
        configuration: T,
    ):
        """
        Initialize self.

        Args:
            configuration: a tracker configuration
        """
        self.configuration = configuration

    @property
    @abstractmethod
    def tracker_type(self) -> str:
        """Get the type of the tracker client."""

    @abstractmethod
    def get_tracker_issue(
        self,
        issue_id: str,
    ) -> Optional[TrackerIssue]:
        """
        Get a tracker issue.

        Args:
            issue_id: an issue id
        """

    @abstractmethod
    def get_tracker_issue_comments(
        self,
        issue_id: str,
        exclude_comments: Optional[List[str]] = None,
    ) -> TrackerIssueComments:
        """
        Get a list of comments on an issue.

        Args:
            issue_id: an issue id
            exclude_comments: an optional list of comment to exclude
        """

    @abstractmethod
    def send_report(
        self,
        report: Report,
    ) -> TrackerIssue:
        """
        Send a report to the tracker.

        Args:
            report: a report
        """

    @abstractmethod
    def send_logs(
        self,
        tracker_issue: TrackerIssue,
        logs: List[Log],
    ) -> SendLogsResult:
        """
        Send logs to the tracker.

        Args:
            tracker_issue: information about the tracker issue
            logs: a list of comments
        """



"""Models and functions for CLI synchronization."""
import argparse

from ywh2bt.cli.error import CliError
from ywh2bt.cli.listener import CliSynchronizerListener
from ywh2bt.core.configuration.error import BaseAttributeError
from ywh2bt.core.core import load_configuration
from ywh2bt.core.factories.tracker_clients import TrackerClientsFactory
from ywh2bt.core.factories.yeswehack_api_clients import YesWeHackApiClientsFactory
from ywh2bt.core.synchronizer.error import SynchronizerError
from ywh2bt.core.synchronizer.synchronizer import Synchronizer


def synchronize(
    args: argparse.Namespace,
) -> None:
    """
    Synchronize data between YWH and trackers.

    Args:
        args: command line arguments

    Raises:
        CliError: if an error occurred
    """
    configuration = load_configuration(
        config_format=args.config_format,
        config_file=args.config_file,
    )
    try:
        configuration.validate()
    except BaseAttributeError as validation_error:
        raise CliError("Invalid configuration") from validation_error
    synchronizer = Synchronizer(
        configuration=configuration,
        yes_we_hack_api_clients_factory=YesWeHackApiClientsFactory(),
        tracker_clients_factory=TrackerClientsFactory(),
        listener=CliSynchronizerListener(),
    )
    try:
        synchronizer.synchronize()
    except SynchronizerError as synchronization_error:
        raise CliError("Synchronization error") from synchronization_error



"""Models used for the configuration of a Jira tracker."""
from typing import (
    Any,
    Optional,
    Text,
)

from ywh2bt.core.configuration.attribute import (
    Attribute,
    BoolAttributeType,
    StrAttributeType,
)
from ywh2bt.core.configuration.tracker import TrackerConfiguration
from ywh2bt.core.configuration.validator import (
    not_blank_validator,
    url_validator,
)


class JiraConfiguration(TrackerConfiguration):
    """A class describing the configuration of a Jira tracker."""

    url: StrAttributeType = Attribute.create(
        value_type=str,
        short_description="API URL",
        description="Base URL of the Jira server",
        required=True,
        validator=url_validator,
    )
    login: StrAttributeType = Attribute.create(
        value_type=str,
        short_description="Login",
        description="User login for the Jira server",
        required=True,
        secret=False,
        validator=not_blank_validator,
    )
    password: StrAttributeType = Attribute.create(
        value_type=str,
        short_description="Password",
        description="User password or API token for the Jira server",
        required=True,
        secret=True,
        validator=not_blank_validator,
    )
    project: StrAttributeType = Attribute.create(
        value_type=str,
        short_description="Project slug",
        description="Jira slug",
        required=True,
        validator=not_blank_validator,
    )
    verify: BoolAttributeType = Attribute.create(
        value_type=bool,
        short_description="Verify SSL",
        description="Verify SSL certs",
        default=True,
    )
    issuetype: StrAttributeType = Attribute.create(
        value_type=str,
        short_description="Issue type",
        description="Issue type (sensitive to account language)",
        default="Task",
    )
    issue_closed_status: StrAttributeType = Attribute.create(
        value_type=str,
        short_description="Issue closed status",
        description="Issue closed status (sensitive to account language)",
        default="Closed",
    )

    def __init__(
        self,
        url: Optional[Text] = None,
        login: Optional[Text] = None,
        password: Optional[Text] = None,
        project: Optional[Text] = None,
        verify: Optional[bool] = None,
        issuetype: Optional[Text] = None,
        issue_closed_status: Optional[Text] = None,
        **kwargs: Any,
    ):
        """
        Initialize the configuration.

        Args:
            url: a Jira API URL
            login: a Jira API user login
            password: a Jira API user password
            project: a Jira project name
            verify: a flag indicating whether to check SSL/TLS connection
            issuetype: a Jira issue type
            issue_closed_status: a Jira name when the issue is closed
            kwargs: keyword arguments
        """
        super().__init__(**kwargs)
        self.url = url
        self.login = login
        self.password = password
        self.project = project
        self.verify = verify
        self.issuetype = issuetype
        self.issue_closed_status = issue_closed_status


TrackerConfiguration.register_subtype(
    subtype_name="jira",
    subtype_class=JiraConfiguration,
)



"""Models and functions used for data synchronisation between YesWeHack and GitHub trackers."""
import os
import re
from copy import deepcopy
from datetime import (
    datetime,
    timezone,
)
from string import Template
from typing import (
    Dict,
    Iterable,
    List,
    Optional,
    Tuple,
    cast,
)

import requests
from github import (
    Github,
    GithubException,
)
from github.Issue import Issue
from github.IssueComment import IssueComment
from github.Repository import Repository

from ywh2bt.core.api.formatter.markdown import ReportMessageMarkdownFormatter
from ywh2bt.core.api.models.report import (
    Attachment,
    Log,
    Report,
)
from ywh2bt.core.api.tracker import (
    SendLogsResult,
    TrackerAttachment,
    TrackerClient,
    TrackerClientError,
    TrackerIssue,
    TrackerIssueComment,
)
from ywh2bt.core.api.trackers.github.attachment import GitHubAttachmentUploader
from ywh2bt.core.configuration.trackers.github import GitHubConfiguration


_RE_IMAGE = re.compile(pattern=r"!\[([^\]]+)]\(([^)]+)\)")
_RE_CONTENT_DISPOSITION_FILENAME = re.compile(pattern='filename="([^"]+)";?')
_TEXT_MAX_SIZE = 65536

AttachmentUploadResult = Tuple[Attachment, Optional[str], Optional[str]]


class GitHubTrackerClientError(TrackerClientError):
    """A GitHub tracker client error."""


class GitHubTrackerClient(TrackerClient[GitHubConfiguration]):
    """A GitHub tracker client."""

    _attachment_name_regex_template = Template(r"(?:!?\[)([^\[\]]*)(?:\])(?:\(${url}\))")
    _attachment_substitute_regex_template = Template(r"!?\[[^\[\]]*\]\(${url}\)")
    _github: Github
    _attachment_uploader: GitHubAttachmentUploader
    _message_formatter: ReportMessageMarkdownFormatter

    _default_timezone: timezone = timezone.utc

    def __init__(
        self,
        configuration: GitHubConfiguration,
    ):
        """
        Initialize self.

        Args:
            configuration: a GitHub configuration
        """
        super().__init__(
            configuration=configuration,
        )
        self._github = Github(
            base_url=cast(str, self.configuration.url),
            login_or_token=cast(str, self.configuration.token),
            verify=cast(bool, self.configuration.verify),
        )
        self._attachment_uploader = GitHubAttachmentUploader(
            configuration=self.configuration,
        )
        self._message_formatter = ReportMessageMarkdownFormatter()

    @property
    def tracker_type(self) -> str:
        """
        Get the type of the tracker client.

        Returns:
            the type of the  tracker client
        """
        return "GitHub"

    def _build_tracker_issue(
        self,
        issue_id: str,
        issue_url: str,
        closed: bool,
    ) -> TrackerIssue:
        return TrackerIssue(
            tracker_url=cast(str, self.configuration.url),
            project=cast(str, self.configuration.project),
            issue_id=issue_id,
            issue_url=issue_url,
            closed=closed,
        )

    def get_tracker_issue(
        self,
        issue_id: str,
    ) -> Optional[TrackerIssue]:
        """
        Get a tracker issue.

        Args:
            issue_id: an issue id

        Returns:
            The issue if it exists, else None
        """
        self._ensure_auth()
        try:
            github_issue = self._get_github_issue(
                issue_id=issue_id,
            )
        except GitHubTrackerClientError:
            return None
        if github_issue is None:
            return None
        return self._build_tracker_issue(
            issue_id=issue_id,
            issue_url=github_issue.html_url,
            closed=github_issue.closed_at is not None,
        )

    def get_tracker_issue_comments(
        self,
        issue_id: str,
        exclude_comments: Optional[List[str]] = None,
    ) -> List[TrackerIssueComment]:
        """
        Get a list of comments on an issue.

        Args:
            issue_id: an issue id
            exclude_comments: an optional list of comment to exclude

        Returns:
            The list of comments
        """
        self._ensure_auth()
        try:
            github_issue = self._get_github_issue(
                issue_id=issue_id,
            )
        except GitHubTrackerClientError:
            return []
        if not github_issue:
            return []
        return self._extract_comments(
            github_issue=github_issue,
            exclude_comments=exclude_comments,
        )

    def send_report(
        self,
        report: Report,
    ) -> TrackerIssue:
        """
        Send a report to the tracker.

        Args:
            report: a report

        Returns:
            information about the sent report
        """
        self._ensure_auth()
        repository = self._get_repository()
        title = self._message_formatter.format_report_title(
            report=report,
        )
        description = self._message_formatter.format_report_description(
            report=report,
        ) + self._get_attachments_list_description(
            attachments=report.attachments,
        )
        external_description = ""
        description_attachment = None
        if len(description) > _TEXT_MAX_SIZE:
            external_description = description
            description_attachment = self._build_external_description_attachment(
                name=f'report-{report.local_id.replace("#", "")}-description.md',
            )
            report_copy = deepcopy(report)
            report_copy.description_html = (
                "<p>This report description is too large to fit into a GitHub issue. "
                + f'See attachment <a href="{description_attachment.url}">{description_attachment.original_name}</a> '
                + "for more details.</p>"
            )
            description = self._message_formatter.format_report_description(
                report=report_copy,
            ) + self._get_attachments_list_description(
                attachments=[
                    description_attachment,
                    *report.attachments,
                ],
            )
        issue = self._create_issue(
            github_repository=repository,
            title=title,
            body="This issue is being synchronized. Please check back in a moment.",
        )
        description, external_description = self._replace_attachments_references(
            uploads=self._upload_attachments(
                issue=issue,
                attachments=report.attachments,
            ),
            referencing_texts=[
                description,
                external_description,
            ],
        )
        if description_attachment:
            description_attachment.data_loader = lambda: bytes(external_description, "utf-8")
            description = self._replace_attachments_references(
                uploads=self._upload_attachments(
                    issue=issue,
                    attachments=[
                        description_attachment,
                    ],
                ),
                referencing_texts=[
                    description,
                ],
            )[0]
        issue.edit(
            body=description,
        )
        return self._build_tracker_issue(
            issue_id=str(issue.id),
            issue_url=issue.html_url,
            closed=False,
        )

    def _build_external_description_attachment(
        self,
        name: str,
    ) -> Attachment:
        return Attachment(
            attachment_id=0,
            name=name,
            original_name=name,
            mime_type="text/markdown",
            size=0,
            url=f"http://tracker/external/{name}",
            data_loader=lambda: bytes("", "utf-8"),
        )

    def send_logs(
        self,
        tracker_issue: TrackerIssue,
        logs: Iterable[Log],
    ) -> SendLogsResult:
        """
        Send logs to the tracker.

        Args:
            tracker_issue: information about the tracker issue
            logs: a list of comments

        Raises:
            GitHubTrackerClientError: if an error occurs

        Returns:
            information about the sent comments
        """
        self._ensure_auth()
        tracker_comments = SendLogsResult(
            tracker_issue=tracker_issue,
            added_comments=[],
        )
        github_issue = self._get_github_issue(
            issue_id=tracker_issue.issue_id,
        )
        if not github_issue:
            raise GitHubTrackerClientError(f"GitHub issue {tracker_issue.issue_id} not found")
        for log in logs:
            github_comment = self._add_comment(
                github_issue=github_issue,
                log=log,
            )
            tracker_comments.added_comments.append(
                TrackerIssueComment(
                    author=github_comment.user.name or github_comment.user.login,
                    created_at=self._ensure_timezone(
                        dt=github_comment.created_at,
                        tz=self._default_timezone,
                    ),
                    comment_id=str(github_comment.id),
                    body=github_comment.body,
                    attachments={},
                ),
            )
        return tracker_comments

    def _ensure_timezone(
        self,
        dt: datetime,
        tz: timezone,
    ) -> datetime:
        if not dt.tzinfo:
            dt = dt.replace(tzinfo=tz)
        return dt

    def test(
        self,
    ) -> None:
        """
        Test the client.

        Raises:
            GitHubTrackerClientError: if the test failed
        """
        try:
            login = self._github.get_user().login
        except GithubException as e:
            raise GitHubTrackerClientError("Unable to log in with GitHub API client") from e
        if not login:
            raise GitHubTrackerClientError("Unable to log in with GitHub API client")

    def _extract_comments(
        self,
        github_issue: Issue,
        exclude_comments: Optional[List[str]] = None,
    ) -> List[TrackerIssueComment]:
        return [
            self._extract_comment(
                github_comment=github_comment,
            )
            for github_comment in github_issue.get_comments()
            if exclude_comments is None or str(github_comment.id) not in exclude_comments
        ]

    def _extract_comment(
        self,
        github_comment: IssueComment,
    ) -> TrackerIssueComment:
        github_body = github_comment.body
        inline_images = _RE_IMAGE.findall(github_body)
        comment_attachments: Dict[str, TrackerAttachment] = {}
        for _, inline_url in inline_images:
            attachment = self._download_attachment(
                url=inline_url,
            )
            if attachment:
                comment_attachments[inline_url] = attachment
        return TrackerIssueComment(
            created_at=self._ensure_timezone(
                dt=github_comment.created_at,
                tz=self._default_timezone,
            ),
            author=github_comment.user.name or github_comment.user.login,
            comment_id=str(github_comment.id),
            body=github_body,
            attachments=comment_attachments,
        )

    def _download_attachment(
        self,
        url: str,
    ) -> Optional[TrackerAttachment]:
        try:
            response = requests.get(url)
        except requests.RequestException:
            return None
        content_disposition = response.headers.get("Content-Disposition")
        filename = None
        if content_disposition:
            match = _RE_CONTENT_DISPOSITION_FILENAME.search(content_disposition)
            if match:
                filename = match.group(1)
        if not content_disposition or not filename:
            filename = os.path.basename(url)
        return TrackerAttachment(
            filename=filename,
            mime_type=response.headers.get("Content-Type", "text/plain"),
            content=response.content,
        )

    def _get_repository(
        self,
    ) -> Repository:
        project = cast(str, self.configuration.project)
        if project[0] == "/":
            project = project[1:]
        try:
            return self._github.get_repo(
                project,
            )
        except (GithubException, requests.RequestException) as e:
            raise GitHubTrackerClientError(
                f"Unable to get GitHub repository {self.configuration.project}",
            ) from e

    def _create_issue(
        self,
        github_repository: Repository,
        title: str,
        body: str,
    ) -> Issue:
        try:
            return github_repository.create_issue(
                title=title,
                body=body,
            )
        except GithubException as e:
            raise GitHubTrackerClientError(
                f"Unable to create issue for project {self.configuration.project} to GitHub",
            ) from e

    def _add_comment(
        self,
        github_issue: Issue,
        log: Log,
    ) -> IssueComment:
        comment_body = self._message_formatter.format_log(log=log,) + self._get_attachments_list_description(
            attachments=log.attachments,
        )
        external_body = ""
        body_attachment = None
        if len(comment_body) > _TEXT_MAX_SIZE:
            external_body = comment_body
            body_attachment = self._build_external_description_attachment(
                name=f"comment-{log.log_id}-description.md",
            )
            log_copy = deepcopy(log)
            log_copy.message_html = (
                "<p>This comment is too large to fit into a GitHub comment. "
                + f'See attachment <a href="{body_attachment.url}">{body_attachment.original_name}</a> '
                + "for more details.</p>"
            )
            comment_body = self._message_formatter.format_log(log=log_copy,) + self._get_attachments_list_description(
                attachments=[
                    body_attachment,
                    *log.attachments,
                ],
            )
        try:
            github_comment = github_issue.create_comment(
                body="This comment is being synchronized. Please check back in a moment.",
            )
        except GithubException as e:
            raise GitHubTrackerClientError(
                f"Unable to add GitHub comment for issue {github_issue} in project {self.configuration.project}",
            ) from e
        comment_body, external_body = self._replace_attachments_references(
            uploads=self._upload_attachments(
                issue=github_issue,
                attachments=log.attachments,
            ),
            referencing_texts=[
                comment_body,
                external_body,
            ],
        )
        if body_attachment:
            body_attachment.data_loader = lambda: bytes(external_body, "utf-8")
            comment_body = self._replace_attachments_references(
                uploads=self._upload_attachments(
                    issue=github_issue,
                    attachments=[
                        body_attachment,
                    ],
                ),
                referencing_texts=[
                    comment_body,
                ],
            )[0]
        github_comment.edit(
            body=comment_body,
        )
        return github_comment

    def _get_github_issue(
        self,
        issue_id: str,
    ) -> Optional[Issue]:
        github_repository = self._get_repository()
        issue_id_int = int(issue_id)
        try:
            for issue in github_repository.get_issues(state="all"):
                if issue.id == issue_id_int:
                    return issue
        except GithubException as e:
            raise GitHubTrackerClientError(
                f"GitHub issue {issue_id} not found in project {self.configuration.project}",
            ) from e
        return None

    def _replace_attachments_references(
        self,
        uploads: List[AttachmentUploadResult],
        referencing_texts: List[str],
    ) -> List[str]:
        for attachment, upload_url, error_message in uploads:
            if upload_url:
                referencing_texts = [
                    text.replace(
                        attachment.url,
                        upload_url,
                    )
                    for text in referencing_texts
                ]
            elif error_message:
                referencing_texts = [
                    self._substitute_attachment_url(
                        body=text,
                        url=attachment.url,
                        substitution=error_message,
                    )
                    for text in referencing_texts
                ]
        return referencing_texts

    def _upload_attachments(
        self,
        issue: Issue,
        attachments: List[Attachment],
    ) -> List[AttachmentUploadResult]:
        uploads = []
        for attachment in attachments:
            url = None
            error_message = None
            if self.configuration.github_cdn_on:
                url = self._attachment_uploader.upload_attachment(
                    issue=issue,
                    attachment=attachment,
                )
                if not url:
                    error_message = f'(Attachment "{attachment.original_name}" not available due to upload error)'
            else:
                error_message = (
                    f'(Attachment "{attachment.original_name}" not available ' + "due to export script’s configuration)"
                )
            uploads.append(
                (
                    attachment,
                    url,
                    error_message,
                ),
            )
        return uploads

    def _get_attachments_list_description(
        self,
        attachments: List[Attachment],
    ) -> str:
        attachments_lines = []
        if attachments:
            attachments_lines = [
                "",
                "**Attachments**:",
            ]
            for attachment in attachments:
                attachments_lines.append(
                    f"- [{attachment.original_name}]({attachment.url})",
                )
            attachments_lines.append("")
        return "\n".join(attachments_lines)

    def _extract_attachment_name(
        self,
        referencing_texts: List[str],
        attachment: Attachment,
    ) -> str:
        attachment_name_regex = self._attachment_name_regex_template.substitute(
            url=re.escape(attachment.url),
        )
        for text in referencing_texts:
            matches = re.findall(
                attachment_name_regex,
                text,
            )
            if matches:
                return cast(str, matches[0])
        return attachment.original_name

    def _substitute_attachment_url(
        self,
        body: str,
        url: str,
        substitution: str,
    ) -> str:
        attachment_substitute_regex = self._attachment_substitute_regex_template.substitute(
            url=re.escape(url),
        )
        return re.sub(
            attachment_substitute_regex,
            substitution,
            body,
        )

    def _ensure_auth(
        self,
    ) -> None:
        try:
            user_id = self._github.get_user().id
        except GithubException as e:
            raise GitHubTrackerClientError("Unable to authenticate to GitHub") from e
        if not isinstance(user_id, int):
            raise GitHubTrackerClientError("Unable to authenticate to GitHub")



"""Models and functions used in markdown conversions."""
import re
from typing import Dict

from ywh2bt.core.api.models.report import Attachment


_RE_IMAGE = re.compile(pattern=r"\!\[([^\]]+)]\(([^)]+)\)")


def markdown_to_ywh(
    message: str,
    attachments: Dict[str, Attachment],
) -> str:
    """
    Convert standard markdown to YWH markdown.

    Args:
        message: a markdown message
        attachments: a list of attachments

    Returns:
        a YWH markdown
    """
    images = _RE_IMAGE.findall(message)
    for key, value in images:
        attachment = attachments.get(value)
        if not attachment:
            continue
        message = message.replace(
            f"![{key}]({value})",
            f"{{YWH-C{attachment.attachment_id}}}",
        )
    return message



"""Models and functions used for states encryption."""
import base64
import json
from typing import (
    Any,
    Dict,
    Tuple,
)

from ywh2bt.core.crypt.encrypt import Encryptor
from ywh2bt.core.crypt.error import CryptError
from ywh2bt.core.crypt.key import Key
from ywh2bt.core.state.error import StateError
from ywh2bt.core.state.state import State


class StateEncryptor:
    """A simple state encryptor."""

    @classmethod
    def encrypt(
        cls,
        state: State,
        key: str,
    ) -> str:
        """
        Encrypt a state with a key.

        Args:
            state: a state
            key: a cryptographic key

        Raises:
            StateError: if the state could not be encrypted

        Returns:
            The encrypted state.
        """
        encryptor = Encryptor(
            key=Key.from_str(
                key=key,
            ),
        )
        try:
            encrypted_data = encryptor.encrypt(
                message_bytes=cls._as_bytes(
                    state=state,
                ),
            )
        except CryptError as e:
            raise StateError(f"Unable to encrypt state {state}") from e
        encrypted_base64 = base64.b64encode(
            encrypted_data,
        )
        encrypted = encrypted_base64.decode("utf-8")
        return f"[YWH2BT:S:{encrypted}]"

    @classmethod
    def _as_bytes(
        cls,
        state: State,
    ) -> bytes:
        json_string = cls._as_json_string(
            state=state,
        )
        return json_string.encode("utf-8")

    @classmethod
    def _as_json_string(
        cls,
        state: State,
    ) -> str:
        try:
            return json.dumps(
                cls._as_meta(
                    state=state,
                ),
            )
        except (TypeError, ValueError) as e:
            raise StateError(f"Unable to serialize state {state}") from e

    @classmethod
    def _as_meta(
        cls,
        state: State,
    ) -> Tuple[str, Dict[str, Any]]:
        return (
            state.__class__.__name__,
            state.as_dict(),
        )



"""CLI arguments parsing and execution."""
import argparse
import sys
from typing import TYPE_CHECKING

import urllib3

from ywh2bt.cli.commands.convert import convert
from ywh2bt.cli.commands.schema import dump_schema
from ywh2bt.cli.commands.synchronize import synchronize
from ywh2bt.cli.commands.test import test
from ywh2bt.cli.commands.validate import validate
from ywh2bt.core.core import (
    AVAILABLE_FORMATS,
    AVAILABLE_SCHEMA_DUMP_FORMATS,
)
from ywh2bt.core.error import (
    print_error,
    print_error_message,
)
from ywh2bt.core.exceptions import CoreException
from ywh2bt.version import __VERSION__


urllib3.disable_warnings(
    category=urllib3.exceptions.InsecureRequestWarning,
)

DEFAULT_FORMATTER_CLASS = argparse.ArgumentDefaultsHelpFormatter
CONFIGURATION_FORMATS = list(AVAILABLE_FORMATS.keys())
SCHEMA_DUMP_FORMATS = list(AVAILABLE_SCHEMA_DUMP_FORMATS.keys())

if TYPE_CHECKING:
    SubParsersActionType = argparse._SubParsersAction[argparse.ArgumentParser]
else:
    SubParsersActionType = argparse._SubParsersAction


def run(
    *args: str,
) -> None:
    """
    Run the CLI.

    Args:
        args: command line arguments
    """
    if not args:
        args = tuple(sys.argv[1:])
    parsed = _parse_args(*args)
    try:
        parsed.func(parsed)
    except CoreException as e:
        print_error(
            error=e,
        )
        sys.exit(1)
    except KeyboardInterrupt:
        print_error_message(
            message="Interrupted by user.",
        )
        sys.exit(130)
    sys.exit(0)


def _parse_args(
    *args: str,
) -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        formatter_class=DEFAULT_FORMATTER_CLASS,
    )
    parser.add_argument(
        "--version",
        "-V",
        action="version",
        version=f"%(prog)s {__VERSION__} (python {'.'.join(map(str, sys.version_info[:3]))})",  # noqa: C812
    )
    parser.set_defaults(func=lambda _: parser.print_help())

    commands = parser.add_subparsers(dest="command")

    _add_validate_command(
        parent_parser=commands,
    )
    _add_synchronize_command(
        parent_parser=commands,
    )
    _add_test_command(
        parent_parser=commands,
    )
    _add_convert_command(
        parent_parser=commands,
    )
    _add_schema_command(
        parent_parser=commands,
    )

    return parser.parse_args(
        args,
    )


def _add_validate_command(
    parent_parser: SubParsersActionType,
) -> None:
    validate_parser = parent_parser.add_parser(
        formatter_class=DEFAULT_FORMATTER_CLASS,
        name="validate",
        help="Validate configuration file (mandatory fields, data types, ...)",
    )
    _add_config_file_format(
        parser=validate_parser,
    )
    validate_parser.set_defaults(func=validate)


def _add_synchronize_command(
    parent_parser: SubParsersActionType,
) -> None:
    synchronize_parser = parent_parser.add_parser(
        formatter_class=DEFAULT_FORMATTER_CLASS,
        name="synchronize",
        aliases=("sync",),
        help="Execute synchronization",
    )
    _add_config_file_format(
        parser=synchronize_parser,
    )
    synchronize_parser.set_defaults(func=synchronize)


def _add_test_command(
    parent_parser: SubParsersActionType,
) -> None:
    test_parser = parent_parser.add_parser(
        formatter_class=DEFAULT_FORMATTER_CLASS,
        name="test",
        help="Test the connection to the trackers",
    )
    _add_config_file_format(
        parser=test_parser,
    )
    test_parser.set_defaults(func=test)


def _add_convert_command(
    parent_parser: SubParsersActionType,
) -> None:
    convert_parser = parent_parser.add_parser(
        formatter_class=DEFAULT_FORMATTER_CLASS,
        name="convert",
        help="Convert a configuration file from a format to another",
    )
    _add_config_file_format(
        parser=convert_parser,
        format_required=True,
    )
    convert_parser.add_argument(
        "--destination-file",
        "-d",
        dest="destination_file",
        help='path to converted file ; if "-", print to stdout',
        default="-",
        type=str,
    )
    convert_parser.add_argument(
        "--destination-format",
        "-df",
        dest="destination_format",
        help="format to converted file",
        required=True,
        type=str,
        default=argparse.SUPPRESS,
        choices=CONFIGURATION_FORMATS,
    )
    convert_parser.add_argument(
        "--override",
        dest="override",
        help="override destination file if it already exists",
        action="store_true",
    )
    convert_parser.set_defaults(func=convert)


def _add_schema_command(
    parent_parser: SubParsersActionType,
) -> None:
    schema_parser = parent_parser.add_parser(
        formatter_class=DEFAULT_FORMATTER_CLASS,
        name="schema",
        help="Dump a schema of the structure of the configuration files in JSON Schema, markdown or plaintext",
    )
    schema_parser.add_argument(
        "--format",
        "-f",
        dest="output_format",
        help="format",
        type=str,
        required=False,
        default=SCHEMA_DUMP_FORMATS[0],
        choices=SCHEMA_DUMP_FORMATS,
    )
    schema_parser.set_defaults(func=dump_schema)


def _add_config_file(
    parser: argparse.ArgumentParser,
) -> None:
    parser.add_argument(
        "--config-file",
        "-c",
        dest="config_file",
        help="path to configuration file",
        required=True,
        type=str,
        default=argparse.SUPPRESS,
    )


def _add_config_format(
    parser: argparse.ArgumentParser,
    required: bool = False,
) -> None:
    parser.add_argument(
        "--config-format",
        "-f",
        dest="config_format",
        help="format of configuration file",
        type=str,
        required=required,
        default="yaml",
        choices=CONFIGURATION_FORMATS,
    )


def _add_config_file_format(
    parser: argparse.ArgumentParser,
    format_required: bool = False,
) -> None:
    _add_config_file(
        parser=parser,
    )
    _add_config_format(
        parser=parser,
        required=format_required,
    )


if __name__ == "__main__":
    run(
        *sys.argv[1:],
    )



"""Models and functions used for format conversion from jira to markdown."""

import re
from typing import (
    Dict,
    List,
    Match,
)

from typing_extensions import Protocol


MatchStr = Match[str]

# inspired by https://github.com/metysj/jira2md.forked


class _ReplacementProtocol(Protocol):
    def __call__(
        self,
        src: str,
    ) -> str:
        ...


_RE_BLOCKQUOTE = re.compile(pattern=r"^bq\.\s+", flags=re.MULTILINE)
_RE_BOLD = re.compile(pattern=r"(^|\s|_)\*(\S.*)(\S)\*", flags=re.MULTILINE)
_RE_CODE = re.compile(
    pattern=r"\{code(:([a-z]+))?([:|]?(title|borderStyle|borderColor|borderWidth|bgColor|titleBGColor)=.+?)*\}",
)
_RE_CITATION = re.compile(pattern=r"(^|\s)\?\?((?:[^?])+)\?\?")
_RE_COLOR = re.compile(pattern=r"\{color:([^}]+)\}(.*)\{color\}", flags=re.MULTILINE | re.DOTALL)
_RE_HEADER = re.compile(pattern=r"^h([0-6])\.(.*)", flags=re.MULTILINE)
_RE_INSERT = re.compile(pattern=r"(^|\s)\+([^+]*)(\S)\+")
_RE_IMAGE = re.compile(pattern=r"!([^!|]+)(\|[^!]*)?!")
_RE_ITALIC = re.compile(pattern=r"(^|\s|\*)_(\S.*)(\S)_")
_RE_LIST = re.compile(pattern=r"^[ \t]*(\*+)\s+", flags=re.MULTILINE)
_RE_MONOSPACE = re.compile(pattern=r"(^|\s){{([^}]+)}}($|\s)")
_RE_NAMED_LINK = re.compile(pattern=r"\[(.+?)\|((www\.|(https?|ftp|ssh):\/\/)[^\s/$.?#].[^\s]*)\]")
_RE_NUMBERED_LIST = re.compile(pattern=r"^[ \t]*(#+)\s+", flags=re.MULTILINE)
_RE_PANEL = re.compile(pattern=r"\{panel:title=([^}]*)\}\n?(.*?)\n?\{panel\}", flags=re.MULTILINE | re.DOTALL)
_RE_QUOTE = re.compile(pattern=r"\{quote\}(.*)\{quote\}", flags=re.MULTILINE | re.DOTALL)
_RE_SIMPLE_LINK = re.compile(pattern=r"\[((www\.|(https?|ftp|ssh):\/\/)[^\s/$.?#].[^\s]*)\]")
_RE_STRIKETHROUGH = re.compile(pattern=r"(^|\s)-(\S+.*?\S)-($|\s)")
_RE_SUBSCRIPT = re.compile(pattern=r"(^|\s)~([^~]*)(\S)~")
_RE_SUPERSCRIPT = re.compile(pattern=r"(^|\s)\^([^^]*)(\S)\^")
_RE_TABLE_CELL = re.compile(pattern=r"\|[^|]+")
_RE_TABLE_HEADER = re.compile(pattern=r"^[ \t]*((?:\|\|.*?)+\|\|)[ \t]*$", flags=re.MULTILINE)
_RE_TABLE_NO_HEADER = re.compile(pattern=r"^[ \t]*((?:\|{1}[^\|\n]+)+\|)[ \t]*$", flags=re.MULTILINE)


def _replace_blockquote(
    src: str,
) -> str:
    return _RE_BLOCKQUOTE.sub("> ", src)


def _replace_bold(
    src: str,
) -> str:
    return _RE_BOLD.sub(r"\1**\2\3**", src)


def _replace_citation(
    src: str,
) -> str:
    return _RE_CITATION.sub(r"\1*&mdash; \2*", src)


def _replace_color(
    src: str,
) -> str:
    return _RE_COLOR.sub(r'<span style="color:\1" class="text-color-\1">\2</span>', src)


def _replace_header_sub(
    match: MatchStr,
) -> str:
    prefix = "#" * int(match.group(1))
    title = match.group(2)
    return f"{prefix}{title}"


def _replace_header(
    src: str,
) -> str:
    return _RE_HEADER.sub(_replace_header_sub, src)


def _replace_insert(
    src: str,
) -> str:
    return _RE_INSERT.sub(r"\1<ins>\2\3</ins>", src)


def _replace_image(
    src: str,
) -> str:
    return _RE_IMAGE.sub(r"![\1](\1)", src)


def _replace_italic(
    src: str,
) -> str:
    return _RE_ITALIC.sub(r"\1*\2\3*", src)


def _replace_list_sub(
    match: MatchStr,
) -> str:
    prefix = "  " * (len(match.group(1)) - 1)
    return f"{prefix}* "


def _replace_list(
    src: str,
) -> str:
    return _RE_LIST.sub(_replace_list_sub, src)


def _replace_monospace(
    src: str,
) -> str:
    return _RE_MONOSPACE.sub(r"\1`\2`\3", src)


def _replace_named_link(
    src: str,
) -> str:
    return _RE_NAMED_LINK.sub(r"[\1](\2)", src)


def _replace_noformat(
    src: str,
) -> str:
    return src.replace("{noformat}", "```")


def _replace_numbered_list_sub(
    match: MatchStr,
) -> str:
    prefix = "  " * (len(match.group(1)) - 1)
    return f"{prefix}1. "


def _replace_numbered_list(
    src: str,
) -> str:
    return _RE_NUMBERED_LIST.sub(_replace_numbered_list_sub, src)


def _replace_panel(
    src: str,
) -> str:
    return _RE_PANEL.sub(r"\n| \1 |\n| --- |\n| \2 |", src)


def _replace_quote_sub(
    match: MatchStr,
) -> str:
    lines = match.group(1).strip().split("\n")
    joined_lines = "\n> ".join(lines).strip()
    return f"> {joined_lines}\n"


def _replace_quote(
    src: str,
) -> str:
    return _RE_QUOTE.sub(_replace_quote_sub, src)


def _replace_simple_link(
    src: str,
) -> str:
    return _RE_SIMPLE_LINK.sub(r"<\1>", src)


def _replace_strikethrough(
    src: str,
) -> str:
    return _RE_STRIKETHROUGH.sub(r"\1~~\2~~\3", src)


def _replace_subscript(
    src: str,
) -> str:
    return _RE_SUBSCRIPT.sub(r"\1<sub>\2\3</sub>", src)


def _replace_superscript(
    src: str,
) -> str:
    return _RE_SUPERSCRIPT.sub(r"\1<sup>\2\3</sup>", src)


def _replace_table_no_header_sub(
    match: MatchStr,
) -> str:
    lookup_offset = match.start(1) - 1
    breaks_count = 0
    while True:
        space = lookup_offset > 0 and match.string[lookup_offset] in {" ", "\n", "\t"}
        if not (space and breaks_count < 2):
            break
        if match.string[lookup_offset] == "\n":
            breaks_count += 1
        lookup_offset -= 1
    if lookup_offset > 0 and match.string[lookup_offset] == "|":
        return match.group(0).strip()
    row = match.group(1).strip()
    empty_header = _RE_TABLE_CELL.sub("| ", row.strip())
    separator = _RE_TABLE_CELL.sub("| --- ", row.strip())
    return f"{empty_header}\n{separator}\n{row}"


def _replace_table_no_header(
    src: str,
) -> str:
    return _RE_TABLE_NO_HEADER.sub(_replace_table_no_header_sub, src)


def _replace_table_header_sub(
    match: MatchStr,
) -> str:
    header = match.group(1)
    single_barred = header.replace("||", "|")
    separator = _RE_TABLE_CELL.sub("| --- ", single_barred)
    return f"{single_barred}\n{separator}"


def _replace_table_header(
    src: str,
) -> str:
    return _RE_TABLE_HEADER.sub(_replace_table_header_sub, src)


class _Jira2Markdown:
    _replacement_map: Dict[str, str]
    _replacement_functions: List[_ReplacementProtocol]

    def __init__(self) -> None:
        self._replacement_map = {}
        self._replacement_functions = [
            # order matters!
            self._replace_code_block,
            _replace_quote,
            _replace_list,
            _replace_numbered_list,
            _replace_header,
            _replace_bold,
            _replace_italic,
            _replace_image,
            _replace_monospace,
            _replace_citation,
            _replace_insert,
            _replace_superscript,
            _replace_subscript,
            _replace_strikethrough,
            _replace_noformat,
            _replace_simple_link,
            _replace_named_link,
            _replace_blockquote,
            _replace_color,
            _replace_table_no_header,
            _replace_panel,
            _replace_table_header,
            self._apply_replacement_map,
        ]

    def convert(
        self,
        src: str,
    ) -> str:
        for replacement_function in self._replacement_functions:
            src = replacement_function(
                src=src,
            )
        return src

    def _replace_code_block(
        self,
        src: str,
    ) -> str:
        src_len = len(src)
        result = ""
        current_word = ""
        collecting = False
        index = -1
        for i, c in enumerate(src):
            char = c
            if char not in {" ", "\t", "\n", "\r", "\v"} and i < src_len - 1:
                current_word += char
                continue
            if i == src_len - 1:
                current_word += char
                char = ""
            match = _RE_CODE.match(current_word)
            if match:
                language = match.group(2) or ""
                result = f"{result}```{language}{char}"
                collecting = not collecting
                if collecting:
                    index += 1
            else:
                if collecting:
                    key = f"@code_({index})_code@"
                    replacement = self._replacement_map.get(key)
                    value = current_word + char
                    if replacement is not None:
                        replacement += value
                    else:
                        replacement = value
                        result += key
                    self._replacement_map[key] = replacement
                else:
                    result += current_word + char
            current_word = ""
        return result

    def _apply_replacement_map(
        self,
        src: str,
    ) -> str:
        for key, value in self._replacement_map.items():
            src = src.replace(key, value)
        return src


def jira2markdown(
    src: str,
) -> str:
    """
    Convert jira string to markdown.

    Args:
        src: a jira string

    Returns:
        jira string converted to markdown
    """
    return _Jira2Markdown().convert(
        src=src,
    )



from typing import (
    List,
    Union,
)


def _sizeof_fmt(
    num: Union[int, float], base: int, units: List[str], max_unit: str, precision: int = 1, suffix: str = "B"
) -> str:
    for unit in units:
        if abs(num) < base:
            return f"{num:.{precision}f}{unit}{suffix}"
        num /= base
    return f"{num:.{precision}f}{max_unit}{suffix}"


def sizeof_fmt_iec(num: Union[int, float], precision: int = 1, suffix: str = "B") -> str:
    return _sizeof_fmt(
        num,
        base=1024,
        units=["", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"],
        max_unit="Yi",
        precision=precision,
        suffix=suffix,
    )


def sizeof_fmt_si(num: Union[int, float], precision: int = 1, suffix: str = "B") -> str:
    return _sizeof_fmt(
        num,
        base=1000,
        units=["", "k", "M", "G", "T", "P", "E", "Z"],
        max_unit="Y",
        precision=precision,
        suffix=suffix,
    )



"""Models and functions used for Yaml data SerDe."""
from typing import (
    Any,
    TextIO,
)

from ruamel.yaml import YAML

from ywh2bt.core.serde import (
    SerDe,
    SerDeError,
)


class YamlSerDe(SerDe):
    """A Yaml Serializer/deserializer."""

    def serialize_to_stream(
        self,
        data: Any,
        stream: TextIO,
    ) -> None:
        """
        Write serialized data as Yaml into a stream.

        Args:
            data: data
            stream: a stream

        Raises:
            SerDeError: if an error occurred during writing or serialization
        """
        yaml = YAML()
        try:
            yaml.dump(
                data=data,
                stream=stream,
            )
        except Exception as e:
            raise SerDeError("YAML dump error") from e

    def deserialize_from_stream(
        self,
        stream: TextIO,
    ) -> Any:
        """
        Read and deserialize Yaml data from a stream.

        Args:
            stream: a stream

        Returns:
            the deserialized data

        Raises:
            SerDeError: if an error occurred during reading or deserialization
        """
        yaml = YAML()
        try:
            return yaml.load(stream.read())
        except Exception as e:
            raise SerDeError("YAML load error") from e



"""Models and functions used for encryption."""
from ywh2bt.core.crypt.error import CryptError
from ywh2bt.core.crypt.key import Key


class Encryptor:
    """A simple encryptor."""

    _key: Key

    def __init__(
        self,
        key: Key,
    ):
        """
        Initialize self.

        Args:
            key: a cryptographic key

        Raises:
            CryptError: if an error occurred
        """
        if not isinstance(key, Key):
            raise CryptError(f"Key should be {Key}. {type(key)} given.")
        self._key = key

    def encrypt(
        self,
        message_bytes: bytes,
    ) -> bytes:
        """
        Encrypt a message.

        Args:
            message_bytes: a message

        Raises:
            CryptError: if an error occurred

        Returns:
            The encrypted message.
        """
        if not isinstance(message_bytes, bytes):
            raise CryptError(f"Message should be {bytes}. {type(message_bytes)} given.")
        key_generator = self._key.generator()
        message_size = len(message_bytes)
        return bytes([message_bytes[i] ^ next(key_generator) for i in range(message_size)])



"""Models and functions used for root configurations GUI."""
from __future__ import annotations

from typing import (
    Any,
    List,
    Optional,
    Tuple,
    cast,
)

from PySide6.QtCore import (
    QFileInfo,
    QSettings,
    Qt,
    Signal,
)
from PySide6.QtWidgets import (
    QBoxLayout,
    QFileDialog,
    QInputDialog,
    QMessageBox,
    QTabWidget,
    QWidget,
)

from ywh2bt.core.core import AVAILABLE_FORMATS
from ywh2bt.core.exceptions import CoreException
from ywh2bt.gui.dialog.error import ErrorDialogMixin
from ywh2bt.gui.dialog.file import FileFormatDialogFilters
from ywh2bt.gui.widgets.root_configuration_entry import RootConfigurationEntry
from ywh2bt.gui.widgets.root_configuration_widget import RootConfigurationWidget
from ywh2bt.gui.widgets.typing import as_signal_instance


class RootConfigurationsWidget(QWidget, ErrorDialogMixin):
    """
    Root configurations GUI.

    Display and edit multiple configurations in visual tabs.
    """

    entry_modified: Signal = Signal(RootConfigurationEntry)
    entry_saved: Signal = Signal(RootConfigurationEntry)
    current_entry_changed: Signal = Signal(RootConfigurationEntry)

    _file_format_dialog_filters: FileFormatDialogFilters
    _settings: QSettings
    _empty_widget: QWidget
    _tab_widget: QTabWidget

    _entries: List[RootConfigurationEntry]

    def __init__(
        self,
        settings: QSettings,
        empty_widget: QWidget,
        parent: Optional[QWidget] = None,
        *args: Any,
        **kwargs: Any,
    ):
        """
        Initialize self.

        Args:
            settings: application settings
            empty_widget: a widget to display if no configurations are currently being edited
            parent: a parent widget
            args: extra arguments
            kwargs: extra keyword arguments
        """
        super().__init__(
            parent,
            *args,
            **kwargs,
        )
        self._settings = settings
        self._entries = []
        self._file_format_dialog_filters = FileFormatDialogFilters(
            formats=AVAILABLE_FORMATS,
        )
        self._init_ui(
            empty_widget=empty_widget,
        )

    def _init_ui(
        self,
        empty_widget: QWidget,
    ) -> None:
        layout = QBoxLayout(
            QBoxLayout.Direction.LeftToRight,
            self,
        )

        self._empty_widget = empty_widget
        self._tab_widget = self._create_tab_widget()

        self._auto_show_hide_tabs()

        layout.addWidget(self._empty_widget, Qt.AlignmentFlag.AlignCenter)
        layout.addWidget(self._tab_widget)
        layout.setContentsMargins(0, 0, 0, 0)

    def _create_tab_widget(
        self,
    ) -> QTabWidget:
        widget = QTabWidget(self)
        widget.setTabsClosable(True)
        as_signal_instance(widget.currentChanged).connect(
            self._on_current_tab_changed,
        )
        as_signal_instance(widget.tabCloseRequested).connect(
            self._on_tab_close_requested,
        )
        return widget

    def _on_current_tab_changed(
        self,
        tab_index: int,
    ) -> None:
        entry = None
        if tab_index >= 0:
            entry = self._entries[tab_index]
        as_signal_instance(self.current_entry_changed).emit(
            entry,
        )

    def _on_tab_close_requested(
        self,
        tab_index: int,
    ) -> None:
        entry = self._entries[tab_index]
        if not entry.has_changed():
            self._on_entry_close_accepted(
                entry=entry,
            )
            return
        reply = QMessageBox.question(
            self,
            f"Close {entry.name}",
            "This configuration has unsaved changes. Close anyway?",
            QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
            QMessageBox.StandardButton.No,
        )
        if reply == QMessageBox.StandardButton.Yes:
            self._on_entry_close_accepted(
                entry=entry,
            )

    def _on_entry_close_accepted(
        self,
        entry: RootConfigurationEntry,
    ) -> None:
        tab_index = self._get_tab_index_for_entry(
            entry=entry,
        )
        self._tab_widget.removeTab(tab_index)
        self._entries.pop(tab_index)
        self._auto_show_hide_tabs()

    def _auto_show_hide_tabs(
        self,
    ) -> None:
        if self._tab_widget.count() == 0:
            self._tab_widget.hide()
            self._empty_widget.show()
        else:
            self._tab_widget.show()
            self._empty_widget.hide()

    def create_new_configuration(
        self,
    ) -> None:
        """Create a new configuration."""
        selected_format, ok = QInputDialog().getItem(
            self,
            "Select format",
            "Format:",
            list(AVAILABLE_FORMATS.keys()),
            0,
            False,
        )
        if not ok:
            return
        self.add_entry(
            entry=RootConfigurationEntry(
                name="New configuration",
                raw="",
                original_raw="",
                raw_format=selected_format,
            ),
        )

    def prompt_open_configuration(
        self,
    ) -> None:
        """Prompt for opening an existing configuration."""
        open_file = self._prompt_open_file()
        if not open_file:
            return
        file_path, file_format = open_file
        self._try_open_configuration(
            file_info=QFileInfo(file_path),
            file_format=file_format,
        )

    def reload_current_configuration(
        self,
    ) -> None:
        """Reload the current configuration."""
        entry = self.get_current_entry()
        if not entry.file:
            return
        reply = QMessageBox.question(
            self,
            f"Reload {entry.name}",
            f"Reload {entry.file.info.filePath()} from disk?\nUnsaved changes will be lost.",
            QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
            QMessageBox.StandardButton.No,
        )
        if reply == QMessageBox.StandardButton.No:
            return
        self._try_open_configuration(
            file_info=entry.file.info,
            file_format=entry.file.file_format,
        )

    def _prompt_open_file(
        self,
    ) -> Optional[Tuple[str, str]]:
        last_opened_dir = self._get_last_opened_dir()
        file_path, chosen_filter = QFileDialog.getOpenFileName(
            self,
            "Open file",
            last_opened_dir or "",
            self._file_format_dialog_filters.get_filters_string(),
        )
        if not file_path:
            return None
        chosen_format = self._file_format_dialog_filters.get_format_name(
            filter_string=chosen_filter,
        )
        if chosen_format is None:
            return None
        self._settings.setValue("last_opened_file", file_path)
        self._settings.setValue("last_opened_file_format", chosen_format)
        return file_path, chosen_format

    def _try_open_configuration(
        self,
        file_info: QFileInfo,
        file_format: str,
    ) -> None:
        try:
            entry = RootConfigurationEntry.from_file(
                file_info=file_info,
                file_format=file_format,
            )
        except CoreException as e:
            self.show_exception_dialog(
                parent=self,
                text="Open error",
                informative_text=f"Unable to open configuration {file_info.filePath()}",
                exception=e,
            )
            return
        self.add_entry(
            entry=entry,
        )

    def try_open_configuration_file(
        self,
        file_path: str,
        file_format: str,
    ) -> None:
        return self._try_open_configuration(
            file_info=QFileInfo(file_path),
            file_format=file_format,
        )

    def _get_last_opened_dir(
        self,
    ) -> Optional[str]:
        last_opened_file = self._settings.value("last_opened_file")
        if not last_opened_file:
            return None
        return (
            QFileInfo(
                cast(str, last_opened_file),
            )
            .dir()
            .path()
        )

    def add_entry(
        self,
        entry: RootConfigurationEntry,
    ) -> None:
        """
        Add a configuration entry to be edited.

        Args:
            entry: a configuration entry
        """
        existing_entry = self._get_same_file_entry(
            entry=entry,
        )
        if existing_entry:
            existing_entry.configuration = entry.configuration
            tab_index = self.refresh_entry(
                entry=existing_entry,
            )
        else:
            widget = RootConfigurationWidget(
                settings=self._settings,
                parent=self,
                entry=entry,
            )
            as_signal_instance(widget.root_configuration_modified).connect(
                self._on_entry_modified,
            )
            as_signal_instance(widget.root_configuration_saved).connect(
                self._on_entry_saved,
            )
            as_signal_instance(widget.root_configuration_saved_as).connect(
                self._on_entry_saved_as,
            )
            self._entries.append(entry)
            tab_index = self._tab_widget.addTab(
                widget,
                _tab_title(
                    entry=entry,
                ),
            )
            self._update_tab_text(
                entry=entry,
            )
        self._tab_widget.setCurrentIndex(tab_index)
        self._auto_show_hide_tabs()

    def has_changed_entries(
        self,
    ) -> bool:
        """
        Check if any opened entry has changed.

        Returns:
            True if any entry has changed.
        """
        return any(entry.has_changed() for entry in self._entries)

    def refresh_entry(
        self,
        entry: RootConfigurationEntry,
    ) -> int:
        """
        Refresh the display of a configuration entry.

        Args:
            entry: a configuration entry

        Returns:
            The index at which this configuration is placed in the edited list
        """
        tab_index = self._get_tab_index_for_entry(
            entry=entry,
        )
        widget = cast(RootConfigurationWidget, self._tab_widget.widget(tab_index))
        widget.set_entry(
            entry=entry,
        )
        self._update_tab_text(
            entry=entry,
        )
        as_signal_instance(self.entry_modified).emit(
            entry,
        )
        return tab_index

    def save_current_entry(
        self,
    ) -> None:
        """Save the current entry configuration."""
        self._get_current_entry_widget().save()

    def save_as_current_entry(
        self,
    ) -> None:
        """Save the current entry configuration in a new file."""
        self._get_current_entry_widget().save_as()

    def test_current_entry(
        self,
    ) -> None:
        """Synchronize the current entry configuration."""
        self._get_current_entry_widget().test()

    def synchronize_current_entry(
        self,
    ) -> None:
        """Synchronize the current entry configuration."""
        self._get_current_entry_widget().synchronize()

    def _get_current_entry_widget(
        self,
    ) -> RootConfigurationWidget:
        return cast(
            RootConfigurationWidget,
            self._tab_widget.widget(
                self._tab_widget.currentIndex(),
            ),
        )

    def _get_same_file_entry(
        self,
        entry: RootConfigurationEntry,
    ) -> Optional[RootConfigurationEntry]:
        if not entry.file:
            return None
        for existing_entry in self._entries:
            if existing_entry.file and existing_entry.file.info == entry.file.info:
                return existing_entry
        return None

    def _on_entry_modified(
        self,
        entry: RootConfigurationEntry,
    ) -> None:
        self._update_tab_text(
            entry=entry,
        )
        as_signal_instance(self.entry_modified).emit(
            entry,
        )

    def _on_entry_saved(
        self,
        entry: RootConfigurationEntry,
    ) -> None:
        as_signal_instance(self.entry_saved).emit(
            entry,
        )

    def _on_entry_saved_as(
        self,
        file_path: str,
        file_format: str,
    ) -> None:
        self._try_open_configuration(
            file_info=QFileInfo(file_path),
            file_format=file_format,
        )

    def _update_tab_text(
        self,
        entry: RootConfigurationEntry,
    ) -> None:
        color = Qt.GlobalColor.black
        if not entry.configuration or not entry.configuration.is_valid():
            color = Qt.GlobalColor.red
        tab_index = self._get_tab_index_for_entry(
            entry=entry,
        )
        self._tab_widget.tabBar().setTabTextColor(
            tab_index,
            color,
        )
        self._tab_widget.setTabText(
            tab_index,
            _tab_title(
                entry=entry,
            ),
        )

    def get_current_entry(
        self,
    ) -> RootConfigurationEntry:
        """
        Get the currently displayed configuration entry.

        Returns:
            The current configuration entry
        """
        tab_index = self._tab_widget.currentIndex()
        return self._entries[tab_index]

    def _get_tab_index_for_entry(
        self,
        entry: RootConfigurationEntry,
    ) -> int:
        return self._entries.index(entry)


def _tab_title(
    entry: RootConfigurationEntry,
) -> str:
    tab_title = entry.name
    if entry.has_changed():
        tab_title = f"* {tab_title}"
    return tab_title



"""Models for errors."""
from typing import (
    Any,
    Dict,
)

from ywh2bt.core.exceptions import CoreException


class BaseAttributeError(CoreException):
    """A class representing a base attribute error."""

    message: str
    context: Any

    def __init__(
        self,
        message: str,
        context: Any,
    ):
        """
        Initialize the error.

        Args:
            message: a message
            context: a context in which the error occurred
        """
        super().__init__(message)
        self.message = message
        self.context = context

    def __repr__(self) -> str:
        """
        Is called by the `repr()` built-in function to compute the "official" string representation.

        Returns:
            The string representation
        """
        message_repr = repr(self.message)
        context_repr = repr(self.context)
        return f"{self.__class__.__name__}({message_repr}, context={context_repr})"


class MissingAttributeError(BaseAttributeError):
    """A class representing a missing attribute error."""


class InvalidAttributeError(BaseAttributeError):
    """A class representing an invalid attribute error."""


class UnsupportedAttributeError(BaseAttributeError):
    """A class representing an extra attribute error."""

    extra: Dict[str, Any]


class AttributesError(BaseAttributeError):
    """A class representing a list of attribute errors."""

    message: str
    errors: Dict[str, BaseAttributeError]

    def __init__(
        self,
        message: str,
        errors: Dict[str, BaseAttributeError],
        context: Any,
    ):
        """
        Initialize the error.

        Args:
            message: a message
            errors: a list of errors
            context: a context in which the error occurred
        """
        super().__init__(
            message=message,
            context=context,
        )
        self.message = message
        self.errors = errors

    def __repr__(self) -> str:
        """
        Is called by the `repr()` built-in function to compute the "official" string representation.

        Returns:
            The string representation
        """
        message_repr = repr(self.message)
        errors_repr = repr(self.errors)
        return f"{self.__class__.__name__}({message_repr}, errors={errors_repr})"

    def __str__(self) -> str:
        """
        Get the string value.

        Is called by `str(object)` and the built-in functions `format()` and `print()`
        to compute the "informal" or nicely printable string representation.

        Returns:
            The string value
        """
        return f"{self.message}"



"""Functions and models used in validators."""
import re
from collections.abc import Sized
from typing import (
    Any,
    Dict,
    TypeVar,
)

from typing_extensions import Protocol


class ValidatorError(Exception):
    """A class representing a validator error."""


T = TypeVar("T", contravariant=True)


class ValidatorProtocol(Protocol[T]):
    """Protocol for validators."""

    def __call__(
        self,
        value: T,
    ) -> None:
        """
        Validate the given value.

        Args:
            value: the value to be validated
        """
        ...


url_validator_regex = re.compile(
    "^(?:http|ftp)s?://"
    + r"(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|"  # http:// or https://
    + "localhost|"  # domain...
    + r"\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})"  # localhost...
    + r"(?::\d+)?"  # ...or ip
    + r"(?:/?|[/?]\S+)$",  # optional port
    re.IGNORECASE,
)

host_validator_regex = re.compile(
    r"(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|"
    + "localhost|"  # domain...
    + r"\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})"  # localhost...
    + r"(?::\d+)?$",  # ...or ip
    +re.IGNORECASE,  # optional port
)


def url_validator(
    value: str,
) -> None:
    """
    Validate an URL.

    Args:
        value: the URL to be validated

    Raises:
        ValidatorError: if the URL is invalid
    """
    matches = re.match(
        pattern=url_validator_regex,
        string=value,
    )
    if not matches:
        raise ValidatorError(f"{repr(value)} is not a valid url")


def host_validator(
    value: str,
) -> None:
    """
    Validate a Host.

    Args:
        value: the Host to be validated

    Raises:
        ValidatorError: if the Host is invalid
    """
    matches = re.match(
        pattern=host_validator_regex,
        string=value,
    )
    if not matches:
        raise ValidatorError(f"{repr(value)} is not a valid host")


def not_empty_validator(
    value: Any,
) -> None:
    """
    Validate the emptiness.

    Args:
        value: the value to be validated

    Raises:
        ValidatorError: if the value is empty
    """
    if value is not False and not value:
        raise ValidatorError(f"{repr(value)} is empty")


def length_one_validator(
    value: Any,
) -> None:
    """
    Validate that the value has a length of 1.

    Args:
        value: the value to be validated

    Raises:
        ValidatorError: if the value does not have a length of one
    """
    if not isinstance(value, Sized):
        raise ValidatorError(f"{repr(value)} does not have a length")
    length = len(value)
    if length != 1:
        raise ValidatorError(f"{repr(value)} does not have a length of one (length={length})")


def not_blank_validator(
    value: str,
) -> None:
    """
    Validate that the value is not a blank string (only whitespaces).

    Args:
        value: the value to be validated

    Raises:
        ValidatorError: if the value is blank
    """
    if not value.strip():
        raise ValidatorError(f"{repr(value)} is blank")


def dict_has_non_blank_key_validator(
    key: str,
) -> ValidatorProtocol[Dict[Any, Any]]:
    """
    Create a validator that validate that the key exists in the dict and its value is not blank.

    Args:
        key: a key to be validated

    Returns:
        a validator
    """

    def validator(
        value: Dict[Any, Any],
    ) -> None:
        key_value = value.get(key, "")
        if not isinstance(key_value, str) or not key_value.strip():
            raise ValidatorError(f"{repr(value)} must contain non-blank string key {repr(key)}")

    return validator



"""Models and functions used for observing synchronisation between YesWeHack and trackers from the CLI."""
import sys
from datetime import datetime

from singledispatchmethod import singledispatchmethod

from ywh2bt.core.api.models.report import REPORT_STATUS_TRANSLATIONS
from ywh2bt.core.core import write_message
from ywh2bt.core.synchronizer.listener import (
    SynchronizerEndEvent,
    SynchronizerEndFetchReportsEvent,
    SynchronizerEndSendReportEvent,
    SynchronizerEvent,
    SynchronizerListener,
    SynchronizerStartEvent,
    SynchronizerStartFetchReportsEvent,
    SynchronizerStartSendReportEvent,
)
from ywh2bt.core.tester.listener import (
    TesterEndEvent,
    TesterEndTrackerEvent,
    TesterEndYesWeHackEvent,
    TesterEvent,
    TesterListener,
    TesterStartEvent,
    TesterStartTrackerEvent,
    TesterStartYesWeHackEvent,
)


def _print(
    message: str,
    end: str = "\n",
) -> None:
    write_message(
        stream=sys.stdout,
        message=message,
        end=end,
    )


def _print_timestamped(
    message: str,
    end: str = "\n",
) -> None:
    now = datetime.now()
    formatted_now = now.strftime("%Y-%m-%d %H:%M:%S.%f")
    _print(
        message=f"[{formatted_now}] {message}",
        end=end,
    )


class CliSynchronizerListener(SynchronizerListener):
    """A listener that print details about the event on the standard output stream."""

    def on_event(
        self,
        event: SynchronizerEvent,
    ) -> None:
        """
        Receive an event.

        Args:
            event: an event
        """
        self._on_event(event)

    @singledispatchmethod
    def _on_event(
        self,
        event: SynchronizerEvent,
    ) -> None:
        listener_name = self.__class__.__name__
        event_name = event.__class__.__name__
        _print(
            message=f"{listener_name}: Unhandled event {event_name}",
        )

    @_on_event.register
    def _on_start(
        self,
        event: SynchronizerStartEvent,
    ) -> None:
        _print_timestamped(
            message="Starting synchronization:",
        )

    @_on_event.register
    def _on_end(
        self,
        event: SynchronizerEndEvent,
    ) -> None:
        _print_timestamped(
            message="Synchronization done.",
        )

    @_on_event.register
    def _on_start_fetch_reports(
        self,
        event: SynchronizerStartFetchReportsEvent,
    ) -> None:
        yeswehack_name = event.yeswehack_name
        program_slug = event.program_slug
        _print_timestamped(
            message=f'  Processing YesWeHack "{yeswehack_name}": ',
        )
        _print_timestamped(
            message=f'    Fetching reports for program "{program_slug}": ',
            end="",
        )

    @_on_event.register
    def _on_end_fetch_reports(
        self,
        event: SynchronizerEndFetchReportsEvent,
    ) -> None:
        reports_count = len(event.reports)
        _print(
            message=f"{reports_count} report(s)",
        )

    @_on_event.register
    def _on_start_send_report(
        self,
        event: SynchronizerStartSendReportEvent,
    ) -> None:
        tracker_name = event.tracker_name
        report = event.report
        if report.title:
            title_max_len = 32
            title = report.title[:title_max_len] + (report.title[title_max_len:] and "...")
            report_details = f"#{report.report_id} ({title})"
        else:
            report_details = f"#{report.report_id}"
        _print_timestamped(
            message=f'    Processing report {report_details} with "{tracker_name}": ',
            end="",
        )

    @_on_event.register
    def _on_end_send_report(
        self,
        event: SynchronizerEndSendReportEvent,
    ) -> None:
        tracker_issue = event.tracker_issue
        issue_added_comment_count = len(event.issue_added_comments)
        if not tracker_issue or not event.is_existing_issue:
            issue_details = [
                "does not exist anymore",
            ]
        else:
            issue_details = [
                tracker_issue.issue_url,
            ]
            if event.is_created_issue:
                if issue_added_comment_count:
                    issue_details.append("updated")
            else:
                issue_details.append("added")
            if issue_added_comment_count:
                issue_details.append(f"{issue_added_comment_count} comment(s) added")
        report_added_comment_count = len(event.report_added_comments)
        report_details = []
        if report_added_comment_count:
            report_details.append(f"{report_added_comment_count} comment(s) added")
        report_details.append(f'tracking status {"updated" if event.tracking_status_updated else "unchanged"}')
        if event.new_report_status:
            old_status, new_status = event.new_report_status
            old_status_translation = REPORT_STATUS_TRANSLATIONS.get(old_status, "Unknown")
            new_status_translation = REPORT_STATUS_TRANSLATIONS.get(new_status, "Unknown")
            report_details.append(f'status "{old_status_translation}" -> "{new_status_translation}"')
        _print(
            message=" | ".join(
                (
                    f'issue => {" ; ".join(issue_details)}',
                    f'report => {" ; ".join(report_details)}',
                ),
            ),
        )


class CliTesterListener(TesterListener):
    """A listener that print details about the event on the standard output stream."""

    def on_event(
        self,
        event: TesterEvent,
    ) -> None:
        """
        Receive an event.

        Args:
            event: an event
        """
        self._on_event(event)

    @singledispatchmethod
    def _on_event(
        self,
        event: TesterEvent,
    ) -> None:
        listener_name = self.__class__.__name__
        event_name = event.__class__.__name__
        _print(
            message=f"{listener_name}: Unhandled event {event_name}",
        )

    @_on_event.register
    def _on_start(
        self,
        event: TesterStartEvent,
    ) -> None:
        _print_timestamped(
            message="Starting test:",
        )

    @_on_event.register
    def _on_end(
        self,
        event: TesterEndEvent,
    ) -> None:
        _print_timestamped(
            message="Test done.",
        )

    @_on_event.register
    def _on_start_test_yeswehack(
        self,
        event: TesterStartYesWeHackEvent,
    ) -> None:
        yeswehack_name = event.yeswehack_name
        _print_timestamped(
            message=f"  YesWeHack {yeswehack_name}: ",
            end="",
        )

    @_on_event.register
    def _on_end_test_yeswehack(
        self,
        event: TesterEndYesWeHackEvent,
    ) -> None:
        _print(
            message="OK",
        )

    @_on_event.register
    def _on_start_test_tracker(
        self,
        event: TesterStartTrackerEvent,
    ) -> None:
        tracker_name = event.tracker_name
        _print_timestamped(
            message=f"  Tracker {tracker_name}: ",
            end="",
        )

    @_on_event.register
    def _on_end_test_tracker(
        self,
        event: TesterEndTrackerEvent,
    ) -> None:
        _print(
            message="OK",
        )



"""Models and functions for CLI conversion."""
import argparse
import sys
from pathlib import Path
from typing import TextIO

from ywh2bt.cli.error import CliError
from ywh2bt.core.configuration.error import BaseAttributeError
from ywh2bt.core.configuration.root import RootConfiguration
from ywh2bt.core.core import (
    get_root_configuration_loader,
    load_configuration,
)
from ywh2bt.core.loader import (
    LoaderError,
    RootConfigurationLoader,
)


def convert(
    args: argparse.Namespace,
) -> None:
    """
    Convert a configuration file.

    Args:
        args: command line arguments

    Raises:
        CliError: if an error occurred
    """
    configuration = load_configuration(
        config_format=args.config_format,
        config_file=args.config_file,
    )
    try:
        configuration.validate()
    except BaseAttributeError as e:
        raise CliError("Invalid configuration") from e
    destination_loader = get_root_configuration_loader(
        file_format=args.destination_format,
    )
    if args.destination_file == "-":
        _convert_to_stream(
            loader=destination_loader,
            configuration=configuration,
            stream=sys.stdout,
        )
        return

    file_path = Path(args.destination_file)
    if file_path.exists() and not args.override:
        raise CliError(f"Unable to open destination file {file_path} (already exist)")
    _convert_to_file(
        loader=destination_loader,
        configuration=configuration,
        file_path=file_path,
    )


def _convert_to_stream(
    loader: RootConfigurationLoader,
    configuration: RootConfiguration,
    stream: TextIO,
) -> None:
    try:
        loader.save(
            data=configuration,
            stream=stream,
        )
    except LoaderError as loader_error:
        raise CliError("Unable to convert configuration (stdout)") from loader_error


def _convert_to_file(
    loader: RootConfigurationLoader,
    configuration: RootConfiguration,
    file_path: Path,
) -> None:
    try:
        with open(file_path, "w") as f:
            _convert_to_stream(
                loader=loader,
                configuration=configuration,
                stream=f,
            )
    except OSError as open_error:
        raise CliError(f"Unable to open destination file {file_path}") from open_error
